
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionTreeDrawer;

import java.util.*;
import java.awt.*;
import java.io.*;
import net.java.games.jogl.GL;
import net.java.games.jogl.GLDrawable;


import AccordionDrawer.*;
/**
 * An abstract class representing a multi-scale rectangular grid on which 
 * a tree consisting of horizontal and vertical line segments is drawn
 *
 * Fills in the fields of the abstract class AccordionDrawer based on
 * information from the Tree class.
 * 
 * @author  Tamara Munzner, Li Zhang
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.TreeEdge
 * @see     AccordionDrawer.GridCell
 */

//public class AccordionTreeDrawer extends AccordionDrawer {
public abstract class AccordionTreeDrawer extends AccordionDrawer {
    String fname;
    protected float pointoffset[] = new float[2];
    protected int labeloffset[] = new int[2];
    protected int labelbuffer[] = new int[2];
    
    public final float pixelDiv = 4.0f;

	public boolean basePass = false;  // draw the base image, nothing marked, no text
	public boolean groupPass = false; // draw the groups, no text
	public boolean textPass = false; // draw the text
	
    /**
     * framebuffer pixels from last mouseover label drawn
     */
    protected float labelPixels[]; // pixels for drawing the label (if there is one)
    protected LabelBox flashLabelBox = null;
    protected float linePixel[][] = new float[2][]; // vertical/horizontal line pixels (if there is one for each)
    protected int lineLength[] = new int[2]; // length of x/y edges (width is line thickness + 1 on each side)
    protected int linePos[][] = new int[2][2]; // starting x/y point for x/y edges

    protected float[][] pal;
//    protected float drawAngle;
    /**
     * GL rasterpos of last mouseover label drawn
     */
    protected double drawnlabelpos[] = new double[2];
	protected int interactionBoxPos[] = new int[2];
    /** 
     * width and height of last mouseover label drawn
     */
    protected int drawnlabelsize[] = new int[2];
	protected int interactionBoxSize[] = new int[2];

    //how the labels are drawn
    // dealing with variable font size
    public int minFontHeight = 10;
    public int maxFontHeight = 14;
    public int popupFontHeight = 14;

    protected float forestShrinkFactor;
    protected float pickoffset[] = new float[2];
    public int pickFuzz; 
    //    static String fontFile = "BitmapFont/OlduvaiFont";
//    static String fontFile = "/OlduvaiFont";
//    static BitmapFontsFamily bff = new BitmapFontsFamily(fontFile);
    static boolean fontInitialized = false;

    public Tree tree;
    protected TreeNode rootNode;
//    protected TreeEdge rootEdge;
    public static boolean fullyQualified;

    // benchmarking 
    public static int countDrawnFrame;
    public static int countDrawnScene;
    public static int cumFrameTime;

	protected Color objectColor;
	public AccordionTreeDrawer(Tree t, int w, int h) {
	super(w,h);
	
	expandleaves = 1;
	dimbrite = true;
	horiz = true;
	objectColor = new Color(0.0f, 0.0f, 0.0f, 1.0f);
	labeloffset[X] = 2;
	labeloffset[Y] = 4;
	labelbuffer[X] = 10;
	labelbuffer[Y] = 1;
	pointoffset[X] = .1f;
	pointoffset[Y] = .52f;
	pickoffset[0] = .1f;
	pickoffset[1] = .1f;
	pickFuzz = 5; // in pixels
	forestShrinkFactor = -0.3f;

	tree = t;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	rootNode = tree.getRoot();
	initCells();
	t.setLeafSplitLine(splitLine[Y]);
//	buildGrids();
//	attachEdges();
	flashGeom = tree.getRoot();
	flashGeomOld = null;
	doBox = false;
//    drawAngle = 30.0f;
//	init(canvas); // no explicit call for jogl, should be called automatically
    }

	/**
	 * clear the objects initiated in ATD
	 * @see TreeJuxtaposer.TreeJuxtaposer.deleteTree
	 * @see AccordionDrawer.close
	 */

		public void shutdown(){
			super.shutdown();      
			tree=null;
			rootNode=null;
//			rootEdge=null;
			fname=null;	 		
			pointoffset=null;
			labeloffset = null;
			labelbuffer = null;		
			drawnlabelpos = null;	
			drawnlabelsize =null;
			
		 }
 
		 protected void finalize() throws Throwable {
		 
					 try {
						 shutdown();
					 }
					 finally {
//						System.out.println("finally clean ATD");
						 super.finalize();     
					 }
		 }


    /**
     * Initialization done after graphics context has been established.
     * @author Tamara Munzner
     */
		 // super is called by jogl, no need to make explicit call
    public void init() {
        super.init(canvas);
    }

	// get the maxline for some node given children have set minlines
	private int getMaxLineX(ArrayList children)
	{
		int maxLine = bottomSize[X];
		Iterator iter = children.iterator();
		while (iter.hasNext())
		{
			int currNodeMinLine = ((TreeNode)iter.next()).getCell().getMinLine(X);
			if (currNodeMinLine < maxLine)
				maxLine = currNodeMinLine; 
		}
		return maxLine;
	}
	
	// update the children minlines to the maxline of the parent
	private void updateMinLineX(ArrayList children, int newMinLine)
	{
		Iterator iter = children.iterator();
		while (iter.hasNext())
		{
			((TreeNode)iter.next()).getCell().setMinLine(newMinLine, X);
		}
	}
	
	// build all splitLines
	private void buildCells()
	{
		int leafCount = 0;
//		Vector leafVector = new Vector();
		int currY = 0; // for positioning leaves in splitLine Y
		for(TreeNode n = tree.posorderStartNode; n!=null; n=n.posorderNext)
		{
			GridCell c = new GridCell(this);
			n.setCell(c);
			n.computedFrame = -1;
			if (n.isLeaf())
			{  // leaf
				splitLine[X].addCell(c, bottomSize[X]-1);
				splitLine[Y].addCell(c, currY++);
//				leafVector.add(n);
				splitLine[Y].setCullingObject(leafCount++, n);
				n.rightmostLeaf = n.leftmostLeaf = n;
//				leafCount++;
			}
			else
			{ // internal node
//				n.addEdge(new TreeEdge(n, false), Y); // vertical edge
				n.rightmostLeaf = n.lastChild().rightmostLeaf;
				n.leftmostLeaf = n.firstChild().leftmostLeaf;
				int maxLineX = getMaxLineX(n.children);
				updateMinLineX(n.children, maxLineX);
				splitLine[X].addCell(c, maxLineX);
				int minLineY = n.leftmostLeaf.getCell().getMinLine(Y);
				int maxLineY = n.rightmostLeaf.getCell().getMaxLine(Y); 
				splitLine[Y].addCell(c, minLineY, maxLineY);
			}
			if (!n.isRoot())
			{
//				n.addEdge(new TreeEdge(n, true), X); // horizontal edge
			}
			else
			{
				rootNode = n;
//				rootEdge = n.getEdge(Y);
				rootCell = c;
			}
		}
	}

    /**
     * Initializes level 0 GridCells based on Tree tree, assign tree
     * nodes to them. Set {@link #maxUsed maxUsed[]} in directions X
     * and Y.
     *
     *
     * @author   Tamara Munzner, Serdar Tasiran
     * 
     *
     * @see      AccordionDrawer.GridCell
     * @see      AccordionDrawer.Tree
     * @see AccordionDrawer.TreeNode */

    public void initCells() {
	int numLeaves = tree.getLeafCount();
	int treeHeight = tree.getHeight();

	bottomSize[0] = treeHeight; 
	bottomSize[1] = numLeaves;
	initSplitLines(true);
	splitLine[X].setMinStuckValue(0.01f);

	gridDepth = new int[2];
	final double logBase2 = Math.log(2);
	for (int xy = 0; xy < 2; xy++)
	{
		gridDepth[xy] = (int)Math.ceil(Math.log(bottomSize[xy]) / logBase2);
		gridDepth[xy] = gridDepth[xy] <= 0 ? 1 : gridDepth[xy];
	}
	if (AccordionDrawer.debugOutput)
	    System.out.println("gridDepth: " +gridDepth[0] + " " +gridDepth[1] +
		" rows:numLeaves " + numLeaves + " cols:treeHeight " + treeHeight +
		" total nodes " + tree.getTotalNodeCount() + " treeNum " + tree.getKey());

	buildCells();
	if (AccordionDrawer.debugOutput)
	    System.out.println("root cell " + rootNode.getCell());
    }

	// x and y are pixel positions
    public CellGeom pickAttached(int x, int y)
    {
    	return pickGeom(x, y);
    }
    
    public CellGeom pickGeom(int x, int y)
    {
		double xPos = s2w(x, X);
		double yPos = s2w(y, Y);
		double xFuzz = s2w(pickFuzz, X);
		double yFuzz = s2w(pickFuzz, Y);
    	return rootNode.pickDescend(xPos, yPos, xFuzz, yFuzz);
    }

    /** 
     * Create box enclosing the subtree beneath input node/edge. The
     * resulting box is the nearest GridLine boundaries to that point,
     * which are cached so that redraws to erase the box can happen
     * after the lines change position.
     * @param cg TreeNode to act upon, passed in as CellGeom from
     *  AccordionDrawer (where TreeNodes are not known)
     * @see IntersectionBox
     * @author Tamara Munzner */
    public InteractionBox makeBox(CellGeom cg)
    {
		if (null == cg)
			return null;
		TreeNode tn;
		if (cg instanceof TreeNode)
		    tn = (TreeNode)cg;
	    else
	    	return null; // not a node or an edge?
		TreeNode minLeaf = tn.leftmostLeaf;
		TreeNode maxLeaf = tn.rightmostLeaf;
		int[] minLine = {tn.getCell().getMinLine(X), minLeaf.getCell().getMinLine(Y)};
		// maxLine[X] could use any leaf to get rightmost splitLine
		int[] maxLine = {minLeaf.getCell().getMaxLine(X), maxLeaf.getCell().getMaxLine(Y)};
	
		return new InteractionBox(minLine, maxLine, cg, tn.getCell().drawer);
    }



    /**
     * On mouseover, flash the object under the mouse and draw its
     * label at maximum size.
     *
     * @author   Tamara Munzner
     * 
     *
     * @see      AccordionDrawer.GridCell
     * @see      AccordionDrawer.Tree
     * @see      AccordionDrawer.TreeNode
     */
    public void doFlash() {
//    	if (debugOutput)
//        System.out.println("doflash start");
		if (noflash
				|| (flashGeom == flashGeomOld && flashBox == flashBoxOld)
				|| keepDrawing())
			return;
//		if (debugOutput)
//System.out.println("entering doFlash");
		doingFlash = true;
		requestRedraw(); // wraps the flash drawing found in flashDraw
//		gl.glFlush(); // no swap, replaces drawEnd
		doingFlash = false;
//		System.out.println("exiting doflash");
	}
    
    // called from within the display function
    public void flashDraw()
    {
		boolean doubleBufferedOriginal = getDoubleBuffer();
//		System.out.println("doublebufferorigninal: " + doubleBufferedOriginal);
		setDoubleBuffer(false); // deactivate double buffering for
									   // drawing flash, redrawing old flash
		
		Color objCol = (null == flashCol) ? objectColor : flashCol;
		if (doBox && flashGeom != null)
			flashBox = makeBox(flashGeom);

		// order matters: 1) undraw old box 2) dump old pixels
		//		  3) save new pixels 4) draw box
		if (doBox && flashBoxOld != null)
			flashBoxOld.undraw();

		int flashBoxBuffer = getLineThickness()/2 + 2; // added to the top/bottom/sides of each line
		int bufferedWidth = 2 * flashBoxBuffer + getLineThickness();
		boolean drawlabelsreal = drawlabels;
		drawlabels = false;
		
		if (null != flashGeomOld)
		{
			// draw in old flash cell
			TreeNode tno = (TreeNode)flashGeomOld;
			boolean[] teo = {tno.getEdge(X), tno.getEdge(Y)};
			for (int xy = X; xy <= Y; xy++)
			{
				if (!teo[xy]) continue;
				if (xy == X && tno.label.length() > 0)
				{
					if (flashLabelBox != null) 
					{
					// draw back around the old label
					gl.glRasterPos3d(s2w(flashLabelBox.pos(X), X),
						s2w(flashLabelBox.pos(Y), Y), getLabelplane());
					gl.glDrawPixels(flashLabelBox.size(X), flashLabelBox.size(Y),
						GL.GL_RGB, GL.GL_FLOAT, labelPixels);
					
					
					}
				}
				// draw back old highlighted lines
				int length[] = {xy == X ? lineLength[xy] : bufferedWidth,
						xy == X ? bufferedWidth : lineLength[xy]};	
				gl.glRasterPos3d(s2w(linePos[xy][X], X),
					s2w(linePos[xy][Y], Y), getLabelplane());
				gl.glDrawPixels(length[X], // length if horiz, else width
					length[Y], // width if horiz, else length
					GL.GL_RGB, GL.GL_FLOAT, linePixel[xy]);
			} // end for loop
		} // end if (null != flashGeomOld)

		if (null != flashGeom)
		{
			// read old cells, draw in new flash cell

			// read pixels
			TreeNode tn = (TreeNode)flashGeom;
			boolean[] te = {tn.getEdge(X), tn.getEdge(Y)};
			gl.glReadBuffer(GL.GL_FRONT);
			for (int xy = X; xy <= Y; xy++)
			{
				if (!te[xy]) continue;
				float[] posStart = new float[2];
				float[] posEnd = new float[2];
				tn.setPositions(posStart, posEnd, xy == X);
				int[] posStartPix = {w2s(posStart[X],X), w2s(posStart[Y],Y)};
				int[] posEndPix = {w2s(posEnd[X],X), w2s(posEnd[Y],Y)};
				if (xy == X && tn.label.length() > 0) // get the horizontal edge to save the pixels under the new big label
				{
					flashLabelBox = tn.makeLabelBox(popupFontHeight, flashX, flashY, posStart, posEnd);
					drawnlabelsize[X] = flashLabelBox.size(X);
					drawnlabelsize[Y] = flashLabelBox.size(Y);
					labelPixels = new float[drawnlabelsize[0] * 3 * drawnlabelsize[1]];
					// 	read pixels under new label from frontbuffer
					gl.glReadPixels(flashLabelBox.pos(X),
						getWinsize(Y) - flashLabelBox.pos(Y),
						drawnlabelsize[X], drawnlabelsize[Y],
						GL.GL_RGB, GL.GL_FLOAT, labelPixels);
				}
				else if (xy == X)
				{
					flashLabelBox = null;
					labelPixels = null;
				}
				// assert: start < end
				if (xy == X)
				{
//					posStart[Y] = posEnd[Y] = (float)tn.getMidY(); // debug, remove later
					posStartPix[Y] = posEndPix[Y] = w2s(tn.getMidY(),Y);
				}
				else // te is vert, has children since it's not null
				{
					posStartPix[Y] = w2s(tn.getChild(0).getMidY(),Y);
					posEndPix[Y] = w2s(tn.getChild(tn.numberChildren()-1).getMidY(),Y);
				}
				lineLength[xy] = posEndPix[xy] - posStartPix[xy] + 2 * flashBoxBuffer;
				linePos[xy][X] = posStartPix[X] - flashBoxBuffer;
				if (linePos[xy][X] < 0) linePos[xy][X] = 0;
				
				linePos[xy][Y] = posEndPix[Y] + flashBoxBuffer; // note: Y coords reversed!
				if (linePos[xy][Y] > getWinsize(Y)) linePos[xy][Y] = getWinsize(Y);
				
				int length[] = {xy == X ? lineLength[xy] : bufferedWidth, 
						xy == X ? bufferedWidth : lineLength[xy]};
				linePixel[xy] = new float[3*length[X]*length[Y]];
				// read under the new drawing, direction xy
				gl.glReadPixels(linePos[xy][X], getWinsize(Y) - linePos[xy][Y],
						length[X], length[Y],
						GL.GL_RGB, GL.GL_FLOAT, linePixel[xy]);
				flag = true; // first frame flag
			}
			
			// don't draw flash for now, debug
			for (int xy = X; xy <= Y; xy++)
			{
				if (!te[xy]) continue;
				
				if (xy == X)// && flashLabelBox != null)
					// draw in label
					if (flashLabelBox != null)
						tn.drawLabelBig(flashX, flashY, popupFontHeight, xy == X);
//					System.out.println("drawing tree node: " + tn);
				tn.drawInCell(flashCol, interactionplane);
			}
			
		} // finished saving pixels and drawing in flash cells
			
		if (doBox && flashBox != null)
			flashBox.draw(rubberbandColor, flashBoxWidth, interactionplane);

		drawlabels = drawlabelsreal;
		if (doBox)
			flashBoxOld = flashBox;
		flashGeomOld = flashGeom;
		flashXOld = flashX;
		flashYOld = flashY;
		setDoubleBuffer(doubleBufferedOriginal); // no redraw
    }

    /**
	 * Decide what GridCells to enqueue for drawing based on tree topology. We
	 * want to add parents and children of current CellGeom. We should find the
	 * GridCell corresponding to the higher-level TreeEdge associated with a
	 * TreeNode, not the level0 GridCell of the TreeNode itself.
	 * 
	 * @author Tamara Munzner
	 */
   	// TODO: decide whether or not this function is useful anymore
   	//  - we don't enqueue cells for drawing anymore
    public void drawNextFancy(CellGeom g) {
//	TreeNode tn;
//	if (g instanceof TreeEdge) {
//	    tn = ((TreeEdge)g).node;
//	} else if (g instanceof TreeNode) {
//	    tn = (TreeNode)g;
//	} else {
//	    return;
//	} 
//	TreeNode par = tn.parent();
//	TreeLeafGridCell c;
//	if (null != par && null != par.edges[X]) {
//	    c = (TreeLeafGridCell)par.edges[X].getCell();
//	    c.possiblyEnqueue();
//	} 
//	if (null != tn.children) {
//	    Iterator iter = tn.children.iterator();
//	    while (iter.hasNext()) {
//		TreeNode tn2 = (TreeNode) iter.next();
//		if (null != tn2.edges[X]) {
//		    c = (TreeLeafGridCell)tn2.edges[X].getCell();
//		    c.possiblyEnqueue();
//		}
//	    }
//	}
    }

    public TreeSet listOfLeaves = null;
    
	protected void drawPreNewFrame() {
		
	if (!groupPass && !textPass) // ok for no snapshots or for the basePass
	{
		countDrawnFrame = 0;
		countDrawnScene = 0;
		cumFrameTime = 0;
		listOfLeaves = splitLine[Y].makePixelRanges(pixelDiv); // divide into block size / 4
		if (debugOutput)
			System.out.println(" (dpnf) list of leaves created: " + listOfLeaves.size());
		pixelSize = getPixelSize(Y);
//	DebugFrame.setStateMachine(DebugFrame.GL_DEPTH _TEST, true);
		// one of these statements causes crash
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glDepthFunc(GL.GL_LEQUAL);
	}

	if (basePass) // first pass of text output
	{
//		System.out.println("staring new output");
		int xSize = winsize[X];
		int ySize = winsize[Y];

		try
		{
			snapShotWriter.write("%!PS-Adobe-2.0 EPSF-2.0\n" +
			"%%BoundingBox: 0 0 " + xSize + " " + ySize + "\n" +
			"%%Magnification: 1.0000\n" +
			"%%EndComments\n" +
			
			"gsave\n" +
			
			"newpath " +
			"0 " + ySize + " moveto " +
			"0 0 lineto " +
			xSize + " 0 lineto " +
			xSize + " " + ySize + " lineto " +
			"closepath " +
			"clip newpath\n" +
			"1 -1 scale\n" +
			"0 -" + ySize + " translate\n");
		}
		catch (IOException ioe)
		{
			System.out.println("Error: unable to write to file: " + snapShotWriter.toString());
		}
	}
    }

	protected void drawPreContFrame() {
	countDrawnFrame = 0;
    }

	protected void drawPostScene() {
//	drawPostFrame();
	gl.glDisable(GL.GL_DEPTH_TEST); // do flash and everything else on top?
	if (textPass) // end of the last pass
	{
		try
		{
			snapShotWriter.write("grestore\n" +
			"showpage\n");
			takeSnapshot = false;
			textPass = false;
			snapShotWriter.close();
		}
		catch (IOException ioe)
		{
			System.out.println("Error: can not write to file: " + snapShotWriter);
		}

	}
//	commented so now isn't needed, if needed, pass it properly, not a state machine!
//	if (dumpstats) {
//	    stats.println(" scene "+countDrawnScene+" frametimeCum "+cumFrameTime+" sceneTIME "+(now-continueStart));
//	    //stats.flush();
//	}
    }

	protected void drawPostFrame() {
	long now = ignoreProgressive ? 0 : System.currentTimeMillis();			
	countDrawnScene += countDrawnFrame;
	long frameTime = now-dynamicStart;
	cumFrameTime += frameTime;
	if (dumpstats) stats.print(" df "+countDrawnFrame+" frameTIME "+frameTime);
    }

	protected void drawBruteForce() {
	drawPreNewFrame();
	drawBruteForceRecurse(rootNode);
	drawPostScene();
    }

    /** 
     * Draw a tree edge, recurse to draw its children.
     * Ignores highlighting, just for benchmarking.
     * @author Tamara Munzner
     */
   	// never called
    private void drawBruteForceRecurse(TreeNode n) {
    	System.out.println("DBFR");
	if (checktime && (0 == countDrawnFrame%100)) {
	    now = System.currentTimeMillis();			
	    if (now-dynamicStart > dynamicTime) {
		return;
	    }
	}
	n.drawInCell(objectColor, objplane);
	Iterator iter = n.children.iterator();
	while (iter.hasNext()) {
	    TreeNode child = (TreeNode)iter.next();
	    if (null != child) drawBruteForceRecurse(child);
	}
    }



    public Tree getTree() { return tree;}
//    public Canvas getCanvas() { return canvas;}
    public GLDrawable getCanvasDrawable() { return (GLDrawable) canvas; }
    public TreeNode getNodeByKey(int key) { return tree.getNodeByKey(key);}
    public TreeNode getNodeByName(String name) { return tree.getNodeByName(name);}

    public int getLabelBuffer(int xy) {return labelbuffer[xy];}
    public void setLabelBuffer(int buffer) {
	if (buffer < 1) buffer = 1;
	labelbuffer[X] = buffer;
	labelbuffer[Y] = buffer;
	requestRedraw();
    }
    public void setLabelBuffer(int buffer, int xy) {
	if (buffer < 1) buffer = 1;
	labelbuffer[xy] = buffer;
	requestRedraw();
    }
    public void increaseLabelBuffer(int xy) {
	setLabelBuffer(labelbuffer[xy]+1, xy);
    }
    public void increaseLabelBuffer() {
	setLabelBuffer(labelbuffer[X]+1);
    }
    public void decreaseLabelBuffer(int xy) {
	setLabelBuffer(labelbuffer[xy]-1, xy);
    }
    public void decreaseLabelBuffer() {
	setLabelBuffer(labelbuffer[X]-1);
    }

    public int getMaxFontHeight() {return maxFontHeight;}
    public void setMaxFontHeight(int fontheight) {
	if (fontheight < 1) fontheight = 1;
	maxFontHeight = fontheight;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	popupFontHeight = maxFontHeight;
	requestRedraw();
    }
    public void increaseMaxFontHeight() {
	setMaxFontHeight(maxFontHeight+1);
    }
    public void decreaseMaxFontHeight() {
	setMaxFontHeight(maxFontHeight-1);
    }

    public int getMinFontHeight() {return minFontHeight;}
    public void setMinFontHeight(int fontheight) {
	if (fontheight < 1) fontheight = 1;
	minFontHeight = fontheight;
	tree.setNodeFontSize(maxFontHeight, minFontHeight);
	requestRedraw();
    }
    public void increaseMinFontHeight() {
	setMinFontHeight(minFontHeight+1);
    }
    public void decreaseMinFontHeight() {
	setMinFontHeight(minFontHeight-1);
    }

	public void startNewFrameAccordionXDrawer()
	{
		drawnLabels.clear();
	}

	public Color getObjectColor() { return objectColor; }
	public void setObjectColor(Color objectColor) { this.objectColor = objectColor;}

	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#subpixelDraw(AccordionDrawer.GridCell)
	 */
	public void subpixelDraw(GridCell c) {
		// TODO Auto-generated method stub, needed to match accordionDrawer abstract class
	
	}

//   public void setDrawAngle(float angle){
//   	drawAngle = angle;
//   	requestRedraw();
//   }
//
//   public float getDrawAngle(){
//   	return drawAngle;
//   }


	 
	public int leafDrawCount;
	public int internalDrawCount;
	
	static long startRender;
	
	public static int dequeueChunkPerTimeCheck = 20;
	

	
	public void drawFrame() {
		if (debugOutput)
			System.out.println("draw frame: " + ToDrawQ.size());
//		splitLine[Y].drawTicks();
		// check to see if layout is correct
			    
	    // Helpful for debugging queue
		if (startFrame)
		{
			leafDrawCount = internalDrawCount = 0;
//			System.out.println("Draw queue size: " + ToDrawQ.size() + " chunks: " + dequeueChunkPerTimeCheck);
			startRender = System.currentTimeMillis();
//			System.out.println(basePass + " " + groupPass + " " + textPass);
		}
	    
		long now = System.currentTimeMillis();		
		while ((takeSnapshot || 
				ignoreProgressive || 
				now - dynamicStart < dynamicTime) && ToDrawQ.size() > 0)
		{
			for (int i = 0; ToDrawQ.size() > 0 && i < dequeueChunkPerTimeCheck; i++)
			{
				Object obj = ToDrawQ.get(0);
				if (obj instanceof RangeInTree)
				{
					RangeInTree r = (RangeInTree)obj; // range of leaf nodes
					drawRange(r.min, r.max);
				}
				else if (obj instanceof TreeNode)
				{
					TreeNode subTreeRoot = (TreeNode)obj; // treenode that is marked, could have marked children, could be taller than a pixel
//					if (subTreeRoot.getCell().drawer != this)
//					{
//						System.out.println("Wrong ATD: " + subTreeRoot.getCell().drawer + " " + this);
//					}
					subTreeRoot.drawDescend(frameNum, objplane, subTreeRoot.leftmostLeaf.key, subTreeRoot.rightmostLeaf.key);
//					subTreeRoot.drawInCell(getColorsForRange(subTreeRoot.key, subTreeRoot.key), objplane);
					if (!subTreeRoot.isRoot())
						subTreeRoot.parent.drawAscend(frameNum, objplane);
				}
				else
					System.out.println("what am I in the todrawQ: " + ToDrawQ.get(0).getClass().getName());
				if (ToDrawQ.size() < 1)
				    System.out.println("synchronization error with drawing queue");
				else
				    ToDrawQ.remove(0);
			} // end the for loop (done chunks per frame)
//			System.out.println("Check the time");
			now = ignoreProgressive ? 0 : System.currentTimeMillis();
			if (!ignoreProgressive && keepMoving()) // break the loop after first draw in animations
				break;
		}
		
		// Queue debugging
//	    if (ToDrawQ.size() >= 0)
//	    {
//	        System.out.println(leafDrawCount + " leaves, " +
//	                internalDrawCount + " internals, in " + 
//	                (System.currentTimeMillis() - startRender) + " ms");
//			System.out.println("Draw queue size: " + ToDrawQ.size());
//	    }

	    drawPostFrame();

		if (keepMoving())
		{
			ToDrawQ.clear();
			incrementFrameNumber();
			if (debugOutput)
				System.out.println("** keep Moving, clearing queue");
//		    keepDrawing = true;
//			keepDrawing = false;
//			incrementFrameNumber();
			// don't call redraw in a drawing context
//			requestRedrawEventually();
		}
		else if (keepDrawing()) {
			if (debugOutput)
				System.out.println("** queue size > 0 (" + ToDrawQ.size() + ")");
//			setKeepDrawing(true);
//			forcedRedraw = true;
			// don't call redraw in a drawing context
//			requestRedrawEventually();
		} else {
//		    System.out.println("** post scene");
//			setKeepDrawing(false);
//	        System.out.println(leafDrawCount + " leaves, " +
//	                internalDrawCount + " internals, in " + 
//	                (System.currentTimeMillis() - startRender) + " ms");
			if (colorgrid || drawBackground)
				splitLine[Y].drawBackground(null, frameNum);
			drawPostScene();
			startFrame = false;
		}
	}

	
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#drawRange(int, int)
	 *
	 *
	 *
	 *	 */
	// min and max are splitLines indexed in Y, surround leaf cells
	// it's known that the distance between min and max is no more than a pixel
	//  - get the min/max cells, use tree.leaf array
	//   - min is the min index into tree.leaf
	//   - max is the max index into tree.leaf
	//  - find parents of all cells (binary recursion, if min.parent != max.parent, recurse on half)
	// cache drawn frame in each cell so we stop at already drawn ancestor
	
	double pixelSize; // set by drawprenewframe
	
	public void drawRange(int min, int max)
	{
//	    System.out.println("draw Range: " + min + " " + max);
		if (min == max)
		{
			System.out.println("no range");
			return; // no split range
		}
		TreeNode minChild = tree.getLeaf(min+1);
		if (min + 1 == max) // single node in a pixel, no sub pixel drawing
		{
			ArrayList colorsForRange = getColorsForRange(minChild.key, minChild.key);
			minChild.drawInCell(colorsForRange, objplane);
			minChild.getCell().setDrawnFrame(frameNum);
			if (minChild.parent != null)
			    minChild.parent.drawAscend(frameNum, objplane);
			return;
		}

		TreeNode maxChild = tree.getLeaf(max);

		// rendering fix
//		Vector subpixelForest = new Vector(); // store each subtree root in this vector that is larger than a vertical pixel, debugging
		int currMin = minChild.key;
//		double pixelSize = getPixelSize(Y); // div pixelDiv (integer > 2, <=? 4) to cover gaps left if using real pixel size
		
		// remember the left most node for drawing (leftmost is closest to the root on the ancestor side of the node)
		int leftMostPosition = minChild.getCell().getMinLine(X);
		TreeNode leftMostNode = minChild; // init
		
		// remember the left most node that is colored for drawing
		// different from uncolored node since uncolored portions aren't drawn when this node isn't as far left as
		//  some other node for this pixel, but need to draw some colored single pixel node if it isn't the farthest
		//  node to the left
		int leftMostColorPosition = minChild.getCell().getMaxLine(X);
		TreeNode leftMostColorNode = null; // init
		Color leftMostColor = null;
		
		TreeNode currRootNode = getNodeByKey(currMin).parent;
		while (currMin <= maxChild.key)
		{
			// ascend tree to pixelSize
			TreeNode prevRootNode = null; // last subpixel node
			double currCellSize = currRootNode.getCell().getSize(Y);
			double prevCellSize = 0;
			while (currCellSize < pixelSize)
//				&& currRootNode.rightmostLeaf.key <= maxChild.key) // need ranges that start on the edge of cells?
			{
				prevRootNode = currRootNode;
				currRootNode = currRootNode.parent;
				prevCellSize = currCellSize;
				currCellSize = currRootNode.getCell().getSize(Y);
			}
			if (prevRootNode == null) // ok now, since we start currRootNode at one up from leaf
			{
				prevRootNode = currRootNode;
//				System.out.println("something is wrong");
			}
			if (prevRootNode != null)
			{
				// add ascended node, for debugging
//				subpixelForest.add(prevRootNode);
				
				ArrayList colors = getColorsForCellGeom(prevRootNode);
				int minLinePositionCheck = prevRootNode.getCell().getMinLine(X); 
				if (minLinePositionCheck < leftMostPosition)
				{
					leftMostPosition = minLinePositionCheck;
					leftMostNode = prevRootNode;
				}
				if (colors.size() > 0 && 
						(minLinePositionCheck < leftMostColorPosition || 
							leftMostColor.getAlpha() < ((Color)colors.get(0)).getAlpha()))
				{
					leftMostColorPosition = minLinePositionCheck;
					leftMostColorNode = prevRootNode;
					leftMostColor = (Color)colors.get(0);
				}
//				prevRootNode.drawDescend(frameNum, objplane, minChild.key, maxChild.key); // pick any child in the min/max range and draw to the leaves
//				if (!prevRootNode.isRoot())
//					prevRootNode.parent.drawAscend(frameNum, objplane);
			}
			
			// set up next round
			int temp = currRootNode.rightmostLeaf.key+1; // some key in the next tree over
			TreeNode tempNode = getNodeByKey(temp); // the next subtree root
			if (tempNode != null)
			{
				currMin = tempNode.leftmostLeaf.key;
				currRootNode = tempNode.leftmostLeaf;
			}
			else // beyond range of nodes
				currMin = temp;
		}

		// colored node isn't the leftmost?  need to draw in the colored node separately
		if (leftMostColorNode != null && leftMostColorPosition != leftMostPosition)
		{
			leftMostColorNode.drawDescend(frameNum, objplane, minChild.key, maxChild.key);
			if (!leftMostColorNode.isRoot())
				leftMostColorNode.parent.drawAscend(frameNum, objplane);
		}
		// the uncolored choice for this pixel or the colored choice is the same:
		leftMostNode.drawDescend(frameNum, objplane, minChild.key, maxChild.key); // pick any child in the min/max range and draw to the leaves
		if (!leftMostNode.isRoot())
			leftMostNode.parent.drawAscend(frameNum, objplane);
	}

	
	public void addToDrawQueue(Object r)
	{
		if (r != null)
		{
			if (r instanceof RangeInTree)
				ToDrawQ.add(((RangeInTree)r).getSplitLineRange(Y, false));
			else if (r instanceof TreeNode)
				ToDrawQ.add(r);
		}	
	}
	
	public abstract ArrayList getColorsForRange(int min, int max);
	public abstract RangeList getRangeByKey(int key);
	
	public void drawDensity( int minObj, int maxObj)
	{ // only applies to sequences
	}
    
	/**
	 * @return
	 */
	public TreeNode getLeaf(int leafNum) {
		return tree.getLeaf(leafNum);
	}
	
	public void splitRangeHelp(Vector drawQueue, double pixelSize, RangeInTree rit)
	{
	    int half = (rit.min + rit.max) / 2;
	    if ((getLeaf(rit.max).getCell().getMax(Y) - 
		        getLeaf(rit.min).getCell().getMin(Y) < pixelSize) ||
		     (rit.min == rit.max))
	    {	
	        drawQueue.add(rit);
	        return;
	    }
	    RangeInTree min = new RangeInTree(rit.min, half, rit.atd);
	    RangeInTree max = new RangeInTree(half+1, rit.max, rit.atd);
	    splitRangeHelp(drawQueue, pixelSize, min);
	    splitRangeHelp(drawQueue, pixelSize, max);
	}
	
	public void splitRanges(double pixelSize)
	{
	    Vector newDrawQueue = new Vector();
		Iterator iter = ToDrawQ.iterator();
		while (iter.hasNext())
		{
			Object nextObj = iter.next();
			if (nextObj instanceof RangeInTree)
			{
				RangeInTree rit = (RangeInTree)nextObj;
			    splitRangeHelp(newDrawQueue, pixelSize, rit);
			}
			else if (nextObj instanceof TreeNode)
				newDrawQueue.add(nextObj);
		}
		ToDrawQ = newDrawQueue;
	}
	
    public void splitRanges(int interval)
    {
		Vector newDrawQueue = new Vector();
		Iterator iter = ToDrawQ.iterator();
		while (iter.hasNext())
		{
			Object nextObj = iter.next();
			if (nextObj instanceof RangeInTree)
			{
				RangeInTree rit = (RangeInTree)nextObj;
				if (rit.max - rit.min + 1 < interval)
					newDrawQueue.add(rit);
				else
				{
					int i = rit.min;
					for (i = rit.min; i < rit.max - interval; i += interval)
						newDrawQueue.add(new RangeInTree(i, i + interval - 1, rit.atd));
					newDrawQueue.add(new RangeInTree(i, rit.max, rit.atd));
				}
			}
			else if (nextObj instanceof TreeNode)
				newDrawQueue.add(nextObj);
		}
		ToDrawQ = newDrawQueue;
	}

	public int getBottomGrid( boolean horizontal )
	{
		return ( horizontal ? bottomSize[AccordionDrawer.X] : bottomSize[AccordionDrawer.Y]) - 1;	
	}
	
	
    
	/**
	 * @return Returns the listOfLeaves.
	 */
	public TreeSet getListOfLeaves() {
		return listOfLeaves;
	}
    
    public void resetSplitValues()
    {
        // for non-uniform split lines (not this type of layout)
    }
};
