/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package TreeJuxtaposer;
 
import java.util.*;
import java.awt.*;
import javax.swing.*;
import java.io.*;

import AccordionDrawer.*;
import AccordionTreeDrawer.*;
import Parser.status.*;

/**
 * A class representing a tree juxtaposer.
 * 
 * A TreeJuxtaposer takes a set of trees and build the data structures
 * for each pair for best corresponding nodes lookup and range
 * intersection queries.  The TreePairs class encapsulates all the
 * Tree2Tree classes required to relate pairs of trees in the list
 * "trees".
 * 
 *
 * @author Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.CellGeom
 * @see     TreeJuxtaposer.TreePairs
 * @see     AccordionDrawer.AccordionTreeDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 */


public class TreeJuxtaposer extends Object {// implements WindowListener, WindowFocusListener {
    
    public static final String versionNumber = "2.0";
    
    Frame mainFrame;
    String title;
    
    StateFrame stateFrame;
    SettingsFrame settingsFrame;
    IncrementalSearch searchFrame;
    DebugFrame debugFrame;
    ButterflyFrame butterflyFrame;
    JPanel controlPanel;
    Panel drawPanel;
    GridLayout drawLayout;
    
   
    /**
     * Resizeable array of AccordionTreeDrawerFinal objects
     */
    ArrayList treeDrawers; 
    ArrayList ATDMatrix;

    /**
     * Resizeable array of Tree objects
     */
    ArrayList trees;
    
    /**
     * Rearrangable list of RangeLists (groups)
     */
    LinkedList groups; 
    Vector groupsByIndex;

    /**
     * an array of TreePairs objects 
     * each TreePair contains the necessary data structure for
     * each pair
     **/

    // ST: Don't understand this one well. Need help documenting.

    TreePairs TPs;

    Color backgroundColor;
    Color objectColor;
    Color labelColor;
    Color labelBackColor;
    Color labelHiColor;
    Color labelBackHiColor;
//    Color rubberbandColor;


    /** these numbers are permanent, but the ordering of the groups
     * changes to reflect user interactions for which color wins out
     * in a tie
     */
    static int flashGroup;
    static int diffGroup;
    static int lcaGroup;
    static int markGroup[] = new int[4];
    static int foundGroup;
    static int treeCount;

    CellGeom flashGeomOld;
    // for debug
    boolean showStructDiffs;
    boolean matrix;

    int linethickness = 1;

    static Status parser_status = new Status(System.in);

    boolean firstTime = true;
    private int treeArgs;

    boolean quasimode;
    AccordionTreeDrawer wantsFocusATD;

    int edgeweightLevels;
    int edgeweightLevelUsed;
    
    static TreeJuxtaposer self;
    
    UI ui;
    float bcnScore;
    
    // TODO: this belongs in the ATD, not HERE!
//    int viewDirection;
//    int leftRight;
//    public static boolean drawButterFly;
    
    int[] winSize = new int[2];
	public ArrayList lcaNode = new ArrayList();

    
    public TreeJuxtaposer() {


	self = this;
	AccordionTreeDrawer.fullyQualified = false; // if true, all nodes will have fully qualified names
	// fully qualified names are good for trees that look like file systems
	// fully qualified names are bad for trees that look like phylogenies
	showStructDiffs = true;
	matrix = false;
	// we use both swing and awt components,
	// so must disable the swing focus manager!
	FocusManager.disableSwingFocusManager();
//	doUI();
    title="TreeJuxtaposer";
    ui = new UI(this, title);
    mainFrame = ui.getMainFrame();

//    mainFrame.addWindowFocusListener(this);
//    mainFrame.addWindowListener(this);
    
    drawPanel = ui.getDrawPanel();
    controlPanel = ui.getControlPanel();
    searchFrame = ui.getSearchFrame();
    stateFrame = ui.getStateFrame();
    debugFrame = ui.getDebugFrame();
    settingsFrame = ui.getSettingsFrame();   
    butterflyFrame = ui.getButterflyFrame(); 
    
	treeDrawers = new ArrayList();
	ATDMatrix = new	ArrayList(); 
	trees = new ArrayList();
	groups = new LinkedList();
	groupsByIndex = new Vector(7);

	backgroundColor = Color.getHSBColor(0.0f/360f,0f,1f);
	// remember that rbcol is xor'ed against backcolor, 
	// so set to opposite of desired color...
//	rubberbandColor = Color.getHSBColor(0.0f/360f,.0f,0.3f);
	objectColor = Color.getHSBColor(0.0f/360f,.0f,0.15f);
	//labelColor = Color.getHSBColor(0f/360f,.0f,0.15f);
	//labelBackColor = Color.getHSBColor(0.0f/360f,0.0f,0.90f);
	//labelColor = Color.getHSBColor(100f/360f,1f,0.7f);
	labelColor = Color.getHSBColor(0f/360f,0f,0f);
	labelBackColor = Color.getHSBColor(0.0f/360f,0.0f,1f);
	labelHiColor = Color.getHSBColor(0.0f/360f,.0f,0.15f);
	labelBackHiColor = Color.getHSBColor(36f/360f,1f,1f);

	Color flashCol = Color.getHSBColor(36f/360f,1f,1f); //orange
	Color foundCol = Color.getHSBColor(310f/360f,1f,1f); //pink
	Color diffCol = Color.getHSBColor(0f/360f,1f,1f); //red
	Color lcaCol = new Color(87, 152, 64);
	//Color diffCol = new Color(87, 152, 64); // dark green, maybe good for contest?
	
	Color markCol[] = new Color[4];
	markCol[0] = Color.getHSBColor(250f/360f,1f,1f); //blue
	markCol[1] = Color.getHSBColor(140f/360f,1f,1f); //green
	markCol[2] = Color.getHSBColor(180f/360f,1f,1f); //cyan
	markCol[3] = Color.getHSBColor(280f/360f,1f,1f); //magenta

// Original color choices.
//	Color flashCol = new Color(1.0f,.7f,0.0f); // orange
//	Color foundCol = new Color(1.0f,0.0f,1.0f); // magenta
//	Color diffCol = new Color(1.0f,.0f,.0f); //red
//	Color markCol[] = new Color[4];
//	markCol[0] = new Color(.0f,1.0f,.0f); // green
//	markCol[1] = new Color(.0f,.0f,1.0f); // blue
//	markCol[2] = new Color(.0f,1.0f,1.0f); // cyan
//	markCol[3] = new Color(1.0f,1.0f,.0f); // yellow
//	labelColor = new Color(.9f,.9f,.9f); // unused
//	labelBackColor = new Color(0.0f,0.0f,0.0f); // black
//	labelHiColor = new Color(0.0f,0.0f,0.0f); // black
//	labelBackHiColor = new Color(1.0f,.7f,0.0f); // orange
//	objectColor = new Color(.7f,.7f,.7f); // white
//	backgroundColor = new Color(0.0f,0.0f,0.0f); // background black

	TPs = new TreePairs();
	treeCount = 0;

	// the order of adding groups matters: the *first* group found
	// gets precedence for color. so the mouseover
	// highlight trumps everything, then the leaf selection.
	// finally, we want the structural differences to
	// take priority over the manual highlighting.

/*	flashGroup = addGroup(false);
	setGroupColor(flashGroup, flashCol);
	// enabled by default, unset since should be off
	setGroupEnable(flashGroup, false);
	foundGroup = addGroup(false);
	setGroupColor(foundGroup, foundCol);
	diffGroup = addGroup(true);
	setGroupColor(diffGroup, diffCol);
	// order of markgroups can change, with setGroupPriority
	markGroup[0] = addGroup(false);
	setGroupColor(markGroup[0], markCol[0]);
	markGroup[1] = addGroup(false);
	setGroupColor(markGroup[1], markCol[1]);
	markGroup[2] = addGroup(false);
	setGroupColor(markGroup[2], markCol[2]);
	markGroup[3] = addGroup(false);
	setGroupColor(markGroup[3], markCol[3]);
	setGroupPriority(markGroup[3]);
	setGroupPriority(markGroup[2]);
	setGroupPriority(markGroup[1]);
	setGroupPriority(markGroup[0]);
	setShowDiffs(true);
*/
//	fc.addChoosableFileFilter(new NexusFilter());
//	fc.addChoosableFileFilter(new NewickFilter());
//	fc.setMultiSelectionEnabled(true);

// diffs > marked group
	diffGroup = addGroup(true);
	setGroupColor(diffGroup, diffCol);

	markGroup[0] = addGroup(false);
	 setGroupColor(markGroup[0], markCol[0]);
	 markGroup[1] = addGroup(false);
	 setGroupColor(markGroup[1], markCol[1]);
	 markGroup[2] = addGroup(false);
	 setGroupColor(markGroup[2], markCol[2]);
	 markGroup[3] = addGroup(false);
	 setGroupColor(markGroup[3], markCol[3]);

	flashGroup = addGroup(false);
	setGroupColor(flashGroup, flashCol);
	// enabled by default, unset since should be off
	setGroupEnable(flashGroup, false);

	foundGroup = addGroup(false);
	setGroupColor(foundGroup, foundCol);
	lcaGroup = addGroup(false);
	setGroupColor(lcaGroup, lcaCol);
	
	setGroupPriority(markGroup[3]);
	setGroupPriority(markGroup[2]);
	setGroupPriority(markGroup[1]);
	setGroupPriority(markGroup[0]);
	
	setShowDiffs(true);

	quasimode = false;
	wantsFocusATD = null;

	edgeweightLevels = 1;
	edgeweightLevelUsed = 0;
    bcnScore = 1.0f;
//    viewDirection = 1;
//    leftRight=0;

    }

    /*
    void lostFocusInQuasi(AccordionTreeDrawer atd) {
	lostFocusATD = atd;
    }
    */
    void wantsFocusInQuasi(AccordionTreeDrawer atd) {
	wantsFocusATD = atd;
    }

    void setQuasimode(boolean on) {
	if (quasimode == true && on == false && wantsFocusATD != null) {
	    //lostFocusATD.transferFocus();
	    wantsFocusATD.getCanvas().requestFocus();
	}
	quasimode = on;
	//quasiATD = null;
	wantsFocusATD = null;
    }
    boolean getQuasimode() { return quasimode; }

    public void updateTitle() {
        title = "TreeJuxtaposer v" + versionNumber;
        if (trees.size() > 0)
        {
            title += ": ";
            for (int i = 0; i < treeDrawers.size(); i++)
            {
                String currTree = ((Tree)trees.get(i)).getName();
                int lastSlash = Math.max(currTree.lastIndexOf('/'), currTree.lastIndexOf('\\'));
                title += currTree.substring(lastSlash+1) + " "; 
            }
            title = title.substring(0, title.length()-1); // remove trailing space
        }
	mainFrame.setTitle(title);
    }


    /**
     * Adds a new tree to "trees", performs all the pairwise tree
     * comparisons and stores results.
     *
     * @author  Tamara Munzner
     *
     * @see AccordionDrawer.Tree
     * @see TreePairs.addTree
     */

    public void addTree(Tree newTree) {
	newTree.setKey(treeCount);
	treeCount++;

	int height, width;
	int hgap = 8;
	int vgap = 8;
	//System.out.println("ADDTREE " + treeCount + " " + treeCount);
	if(1 ==treeCount) {
//        if(!drawButterFly)
	    {
	    	height = 540; 
	         width = 640;
	    }
//	    else
//	    {
//	    	height = winSize[1];
//	    	width = winSize[0];
//	    }
	    // hardwired assumption of controlPanel placement, must
	    // change if we move controlPanel! 10/10/02 TMM
	    int controlWidth = controlPanel.getWidth();
	    width = (controlWidth > width) ? controlWidth : width;

	    if (matrix) {
		// set this very small so that it doesn't override the
		// small windows later - when loading multiple files
		// from command line, after we set the initial size it
		// doesn't seem to be possible to shrink things
		// afterward.

		height = 10;
		width = 10;
	    }
	    drawPanel.setSize(width, height);
	    mainFrame.setSize(width, height+controlPanel.getHeight());
	    mainFrame.validate();
	} else {
	    int numTrees = (treeArgs > treeCount) ? treeArgs : treeCount;

	    if (matrix) { 

		Toolkit toolkit = Toolkit.getDefaultToolkit();
		Dimension screendim = toolkit.getScreenSize();
		mainFrame.setSize(screendim.width-10, screendim.height-30);
		        
		mainFrame.setLocation(10,30);
		//int cph  = controlPanel.getHeight();
		int cph = 100;
		width = (screendim.width - 20 - hgap*(numTrees-1))/numTrees;
		height = (screendim.height - 30 - cph - vgap*(numTrees-1))/numTrees;
		drawPanel.setSize(screendim.width, screendim.height-cph);
		drawLayout = new GridLayout(0,numTrees,hgap,vgap);
		drawPanel.setLayout( drawLayout );
		drawPanel.validate();
		//System.out.println("scr w " + screendim.width + " scr h " + screendim.height + " w " + width + " h " + height + " cph " + cph);
		//mainFrame.pack();
		mainFrame.validate();
	    } else {
			
		// please set this to 0 here in order to avoid the 'invalid-drawable'
		// problem on macosx
		// I think the problem is, that the canvas set its own w and h 
		// so don't mess with that guy. he is awkward and don't likes
		// sizes other then 0 at startup - Yes I know that this is strange...
		height = 0;// drawPanel.getHeight();
		width = 0;//drawPanel.getWidth()/numTrees;
       
//		drawPanel.setSize(width, height);
//		drawPanel.validate();
		mainFrame.validate();

	    }
	}	

	/**
	 * Construct the data structures for structural comparison and
	 * highlight
	 **/

	trees.add(newTree);
//	TPs.addTree(newTree, edgeweightLevels);

	//newTree.printLeaves();
	addLeaves();

	/**
	 * Create and add the drawer(s) for the tree
	 **/
	if (matrix) {
		TPs.addTree(newTree, edgeweightLevels);
	    // matrix mode is n by n grid, where n is total number of
	    // trees (like scatterplot matrix). there are n rows where
	    // each tree is drawn n times: once with no highlights in
	    // column that corresponds to own row (i.e. the diagonal),
	    // and highlighted by the tree corresponding to that
	    // column. example diagram for three trees: capital letter
	    // is tree drawn, lower case is which tree it's
	    // highlighted against.

            // A   A/b  A/c
            // B/a B    B/c
	    // C/a C/b  C

	    // because of the way the gridlayout works, we can't just
	    // add new components because there's no way to rearrange
	    // or insert at specific locations. so just clear the
	    // panel and then insert everything in the correct order.

	    // reshape before removing, otherwise command is ignored.
	    // nope, this doesn't work either.
	    Iterator tdIter = treeDrawers.iterator();
	    while (tdIter.hasNext()) {
		AccordionTreeDrawer atd = (AccordionTreeDrawer)tdIter.next();
		atd.reshape(atd.getCanvasDrawable(), 10, 10, width,height);
		atd.requestRedraw();
	    }
	    drawPanel.removeAll();
	    ArrayList newRow = new ArrayList();
	    ATDMatrix.add(newRow);
	    for (int row = 0; row < treeCount; row++ ) {
		ArrayList ATDRow = (ArrayList) ATDMatrix.get(row);
		for (int col = 0; col < treeCount; col++) {
		    int index = row*treeCount+col;
		    //System.out.println("row " + row + " col " + col + " ind " + index + " tc " + treeCount + " TDsize " + treeDrawers.size());
		    if (row >= treeCount-1 || col >= treeCount-1) {
			// add new one
			Tree treeToUse;
			if (row == treeCount-1 && 0 == col) {
			    // use new tree for the first new ATD on bottom row
			    treeToUse = newTree;
			    //System.out.println("use tree row " + row + " col " + col);
			} else {
			    // clone the tree, this isn't the first new ATD
			    treeToUse = new Tree(((AccordionTreeDrawer)ATDRow.get(0)).tree);
			    treeToUse.setKey(treeCount);
			    TPs.addTree(treeToUse, edgeweightLevels);
			    //System.out.println("add tree row " + row + " col " + col);
			}
			AccordionTreeDrawer newATD = (AccordionTreeDrawer) addATD(treeToUse, width, height);
			ATDRow.add(newATD);
		    } else {
			// tell old one it got smaller
			//((AccordionTreeDrawer)ATDRow.get(col)).reshape(width,height);
			//((AccordionTreeDrawer)ATDRow.get(col)).requestRedraw();
			//System.out.println("shrink tree row " + row + " col " + col);
		    }
		}
	    }
	    // can't set structural differences until all the new ATDs have been made
	    for (int row = 0; row < treeCount; row++ ) {
		ArrayList ATDRow = (ArrayList) ATDMatrix.get(row);
		for (int col = 0; col < treeCount; col++) {
		    ArrayList ATDCol = (ArrayList) ATDMatrix.get(col);
		    int index = row*treeCount+col;
		    if (showStructDiffs == true) {
			if (row != col && (row >= treeCount-1 || col >= treeCount-1 )) {
				AccordionTreeDrawer atdA = (AccordionTreeDrawer)ATDRow.get(col);
				AccordionTreeDrawer atdB = (AccordionTreeDrawer)ATDCol.get(row);
			    doStructuralDiff(atdA, atdB, diffGroup);

				Tree2Tree t2t = TPs.getPair(atdA.tree, atdB.tree);  // new subtree => forest preprocessing:
				t2t.subtree2Forest(atdA, atdB, edgeweightLevels); //    matrix mode untested
			    //System.out.println("SD row " + row + " col " + col + " ind " + index + " tc " + treeCount);
			}
		    }
		    AccordionTreeDrawer ATD = (AccordionTreeDrawer)ATDRow.get(col);
		    drawPanel.add(ATD.getCanvas());
		    //System.out.println("adding panel for " + row + " " + col + " atdkey " + ATD.getKey() + " treekey " + ATD.tree.getKey());
		    //ATD.requestRedraw();
		}
	    }
	    // create color trees for faster color lookup
		RangeList diffGroupRangeList = getGroup(diffGroup);
	} else {
	    addATD(newTree, width, height);
		TPs.addTree(newTree, edgeweightLevels);
	    if (showStructDiffs == true)
	    {
		for (int i = 0; i < treeDrawers.size()-1; i++) {
			AccordionTreeDrawer atdA = (AccordionTreeDrawer)treeDrawers.get(i);
			AccordionTreeDrawer atdB = (AccordionTreeDrawer)treeDrawers.get(treeCount-1);
		    doStructuralDiff((AccordionTreeDrawer)treeDrawers.get(i), 
				     (AccordionTreeDrawer)treeDrawers.get(treeCount-1), 
				     diffGroup);
		    doStructuralDiff((AccordionTreeDrawer)treeDrawers.get(treeCount-1), 
				     (AccordionTreeDrawer)treeDrawers.get(i), 
				     diffGroup);
			Tree2Tree t2t = TPs.getPair(atdA.tree, atdB.tree); // new subtree => forest preprocessing:
			t2t.subtree2Forest(atdA, atdB, edgeweightLevels);
		}
//		System.out.println("Number of differences: " + ((RangeList)groups.get(diffGroup)).getRanges().size());
//		System.out.println("Difference groups: " + groups.get(diffGroup));
		// create color trees for faster color lookup
//		RangeList diffGroupRangeList = getGroup(diffGroup);
//		diffGroupRangeList.setColorTree(new ColorTree(diffGroupRangeList));
	    }
	    Component c = ((AccordionTreeDrawer)treeDrawers.get(treeCount-1)).getCanvas();
	    drawPanel.add(c);
	}
	
	drawPanel.validate();
	mainFrame.validate();
	mainFrame.repaint();
	drawPanel.repaint();
//	System.out.println("add tree requesting redraw");
//	requestRedrawAll();
 //  drawWithViewDirection();	
    }

// TODO: this all belongs in ATD, in a MUCH better way
//	public void setHorizontalDirection(){
//		viewDirection=1;
//		drawWithViewDirection();
//	}
//
//	public void setVerticalDirection(){
//		viewDirection=0;	
//		drawWithViewDirection();
//	}
//
//   public void drawWithViewDirection()
//   {
//	int k = treeDrawers.size();
//	   String[] file = new String[k];
//	   int[] number = new int[k];
//	   for(int i=0; i<k; i++)
//	   {
//		  file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//		  number[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getNumber();
//	   }
//	   for(int i=0; i<k; i++)
//		 deleteTree(file[i]);
//	   GridLayout newLayout = null;
//	  if(viewDirection==1)
//	  	newLayout = new GridLayout(1,0,8,0);
//	  else
//		newLayout = new GridLayout(0,1,0,8);
//	  drawPanel.setLayout(newLayout);
//	  	for(int i=0; i<k; i++)
//	  	    if (number[i] == 0)
//	  	        loadNewickTree(file[i]);
//	  	    else
//	  	    {
//	  	        Vector v = new Vector();
//	  	        v.add(new Integer(number[i]));
//	  	        loadNexusTree(file[i], v);
//	  	    }
//
//   }			 		  
//   
//// draw Rite Left butterfly layout
//   public void drawRL()
//   {
//   	winSize[0] = this.mainFrame.getSize().width;
//   	winSize[1] = this.mainFrame.getSize().height;
//   	drawButterFly = true;
//	int k = treeDrawers.size();
//		String[] file = new String[k];
//		for(int i=0; i<k; i++)
//		   file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//	    for(int i=0; i<k; i++)
//		  deleteTree(file[i]);  
//
//		 for(int i=0; i<k; i++)
//         {
//         	if(i%2==0)
//         	{
//         		setLeftRight(0);			 
//				loadNewickTree(file[i]);
//         	}
//         	else 
//         	{
//         		setLeftRight(1);				
//				loadNewickTree(file[i]);
//         	}
//         }
////         drawButterFly = false;
//   }
//
//// draw Rite Rite butterfly layout
//   public void drawRR()
//   {
//	winSize[0] = this.mainFrame.getSize().width;
//	winSize[1] = this.mainFrame.getSize().height;
//	drawButterFly = true;
//  	
//	int k = treeDrawers.size();
//	String[] file = new String[k];
//	for(int i=0; i<k; i++)
//	   file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//	for(int i=0; i<k; i++)
//		deleteTree(file[i]);
//	setLeftRight(0);
//	 for(int i=0; i<k; i++)
//		loadNewickTree(file[i]);
////	drawButterFly = false;
//   }
//   
//  // draw Up Up  
//   public void drawUU()
//	 {
//		winSize[0] = this.mainFrame.getSize().width;
//		winSize[1] = this.mainFrame.getSize().height;
//		drawButterFly = true;
//	 	
//	 	int k = treeDrawers.size();
//	 	String[] file = new String[k];
//	 	for(int i=0; i<k; i++)
//	       file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//		for(int i=0; i<k; i++)
//			deleteTree(file[i]);
//	    setLeftRight(2);
//	     for(int i=0; i<k; i++)
//	        loadNewickTree(file[i]);
////	        drawButterFly = false;
//	 }
//   
//	// draw Down Down  
//	  public void drawDD()
//		{
//			winSize[0] = this.mainFrame.getSize().width;
//			winSize[1] = this.mainFrame.getSize().height;
//			drawButterFly = true;
//			
//			int k = treeDrawers.size();
//		   String[] file = new String[k];
//		   for(int i=0; i<k; i++)
//		    file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//		  for(int i=0; i<k; i++)
//		   deleteTree(file[i]);	
//		 setLeftRight(3);
//		 for(int i=0; i<k; i++)
//			loadNewickTree(file[i]);
////			drawButterFly = false;
//		}
//   
//   
//// draw Rite Left butterfly layout
//   public void drawLL()
//   {
//	winSize[0] = this.mainFrame.getSize().width;
//	winSize[1] = this.mainFrame.getSize().height;
//	drawButterFly = true;
//   	
//	int k = treeDrawers.size();
//	String[] file = new String[k];
//	for(int i=0; i<k; i++)
//	   file[i] = ((AccordionTreeDrawer)treeDrawers.get(i)).getTree().getName();
//	for(int i=0; i<k; i++)
//	deleteTree(file[i]);
//	
//	setLeftRight(1);
//	 for(int i=0; i<k; i++)
//		loadNewickTree(file[i]);
//     drawButterFly = false;
//   }
/**
 * draw independent resize window
 *
 */
/*	public void drawIndependentResizeWindow(){
		        int k = treeDrawers.size();
			   JSplitPane[] jsp = new JSplitPane[10];
			   Canvas[]  temp = new Canvas[10];	     
			  JPanel[] jp = new JPanel[10];
				for(int i =0 ; i< jp.length; i++)
				{
				 jp[i] = new JPanel();
				 jp[i].setMinimumSize(new Dimension(10,10));
				 jp[i].setLayout(new BorderLayout());
				} 
			  for(int i=0; i<treeDrawers.size();i++){						
				  temp[i] = (AccordionTreeDrawer)treeDrawers.get(i);
//				  temp[i].setSize(0,0);		
				  jp[i].add(temp[i]);
			  }		  			
		   int wh;
		   if(viewDirection == 1)
			 wh = mainFrame.getSize().width;
			 else
			 wh = mainFrame.getHeight()-30; 
                   
            jsp[0] = new JSplitPane(viewDirection, jp[0], jp[1]);
			 if(k==1)
				 jsp[0].setDividerLocation(wh);  
		    else if(k == 2){
			 	jsp[0] = new JSplitPane(viewDirection, jp[0], jp[1]);
				jsp[0].setDividerLocation(wh/2);
				jsp[0].setContinuousLayout(true);
			 }				                    
             for(int i=2; i<k; i++){
                    jsp[i-2].setDividerLocation(wh*(i-1)/k);
                    jsp[i-2].setContinuousLayout(true);
                 	jsp[i-1] = new JSplitPane(viewDirection, jsp[i-2], jp[i]);
                 	jsp[i-1].setDividerLocation(wh*i/k);
                 	jsp[i-1].setContinuousLayout(true);          
		   }	
		   mainFrame.removeAll();
		   if(k == 0){}
		   else if(k<=2)
			mainFrame.add(jsp[0], BorderLayout.CENTER);		
			else
		    mainFrame.add(jsp[k-2], BorderLayout.CENTER);
//			drawPanel.add(c); 
//			drawPanel.validate();
    		mainFrame.validate();	
//		    mainFrame.repaint();
		  }	
*/


    /**
     * Utility function to make new AccordionTreeDrawer, set initial
     * parameters and do treeDrawer list bookkeeping.
     * @author Tamara Munzner
     */
    private AccordionTreeDrawerFinal addATD(Tree newTree,int width,int height) {
	AccordionTreeDrawerFinal ATD = new AccordionTreeDrawerFinal(newTree, width, height, this);
	ATD.setBackgroundColor(backgroundColor);
	ATD.setObjectColor(objectColor);
	ATD.setLabelColor(labelColor);
	ATD.setLabelBackColor(labelBackColor);
	ATD.setLabelHiColor(labelHiColor);
	ATD.setLabelBackHiColor(labelBackHiColor);
	ATD.setRubberbandColor(Color.getHSBColor(0.0f/360f,.0f,0.3f));
 	ATD.setKey(treeDrawers.size());
	ATD.setLineThickness(linethickness); // no redraw
	treeDrawers.add(ATD);
	ATD.setExpandLeaves(1, false);
//	ATD.setDimColors(true);
//	ATD.drawDirection = getLeftRight();
	return ATD;
    }

	/**
	  * Delete a tree from "trees", clean up all the data structures
	 * constructed for tree comparisons.
	 *
	 * @author  
	 * @see AccordionDrawer.AccordionDrawer.close
	 * @see AccordionDrawer.AccordionTreeDrawer.close
	 * @see AccordionDrawer.GridCell.close
	 * @see AccordionDrawer.Tree.close
	 * @see AccordionDrawer.TreeEdge.close
	 * @see AccordionDrawer.TreeNode.close
	 * @see Parse.Newick.close
	 * @see Tree2Tree.close
	 * @see TreePairs.removeTree
	 */
	public void deleteTrees(int[] treeNums) {
   
		System.out.println("VM memory before operation: "+Runtime.getRuntime().freeMemory()) ;
		System.out.println("Total memory before operation: "+Runtime.getRuntime().totalMemory()) ;
 
		for (int i = treeNums.length-1; i >= 0; i--)
		{
			Tree currTree = (Tree)trees.get(treeNums[i]);
			AccordionTreeDrawer atd = (AccordionTreeDrawer)treeDrawers.get(treeNums[i]);
			if (atd.tree != currTree)
			{
				System.out.println("Error: mismatched tree/drawer in delete: " + currTree + " != " + atd.tree);
				continue;
			}
			Component c = atd.getCanvas();
			TPs.removeTree(currTree);
			currTree.close();
			trees.remove(currTree);
			atd.shutdown();
//			atd.cvsDispose();
			treeDrawers.remove(atd);
			drawPanel.remove(c);
			treeCount--;
		}
		drawPanel.validate();
	}
	
	public void postDeleteTrees()
	{
	  
	  addLeaves();
//	  clearGroup(diffGroup);
	  for (int i=0; i<groups.size(); i++)
	  	clearGroup(i); 
      
	  for (int i = 0; i < treeDrawers.size(); i++)
	  {
	  	AccordionTreeDrawerFinal atdA = (AccordionTreeDrawerFinal)treeDrawers.get(i);
	  	for (int j = i+1; j < treeDrawers.size(); j++)
	  	{
	  		AccordionTreeDrawerFinal atdB = (AccordionTreeDrawerFinal)treeDrawers.get(j);
	  		doStructuralDiff(atdA, atdB, diffGroup);
	  		doStructuralDiff(atdB, atdA, diffGroup);
	  	}
	  }
		System.gc();
		System.out.println("VM memory after operation: "+Runtime.getRuntime().freeMemory()) ;
		System.out.println("Total memory after operation: "+Runtime.getRuntime().totalMemory()) ;       
		changedMarks();
		updateTitle();
		requestRedrawAll();
	}

    /**
     * Fetch a tree by its name.  This is done by linear search as we
     * do not expect to have many trees in memory.
     *
     * @author Li Zhang
     */
    public Tree getTreeByName(String name) {
	
	for(int i=0; i<trees.size(); i++) {
	    Tree t = (Tree)trees.get(i);
	    if(t.getName().equals(name)) return t;
	}
	return null;	
    }
	

    protected void addAction(){
    	ui.addAction();
    }
//	private void quitAction()
    protected void quitAction()
	{
		// add other actions if needed when the quit menu option is selected
		//System.exit(0);
		ui.quitAction();
	}

    public void changedMarks() {
    	Iterator tdIter = treeDrawers.iterator();
	while (tdIter.hasNext()) {
	    AccordionTreeDrawer atd = (AccordionTreeDrawer)tdIter.next();
	    atd.changedMarks();
	    //System.out.println("changedmarks, cache: "+atd.cacherange);
	}
    }
	
    public void requestRedrawAll() {
//try { throw new Exception(); } catch (Exception e) { StackTraceElement[] stack = e.getStackTrace(); System.out.println("RRA    " + stack[1]); }
//        System.out.println("request redraw all");
    	Iterator tdIter = treeDrawers.iterator();
	while (tdIter.hasNext()) {
	    AccordionTreeDrawer atd = (AccordionTreeDrawer)tdIter.next();
	    atd.requestRedraw();
	}

    }
    
    public boolean isLoaded()
    {
    	return AccordionDrawer.loaded;
        
    }


    // resets the leaves and recreates the node hash set and search array
    public void addLeaves()
    {
    	// hash set has constant time verification that a similarly named node has already been added
    	// a tree set is also capable of doing this job, but requires rebalancing
    	HashSet hn;
    	if (trees.size() > 0)
    		hn = new HashSet(((Tree)trees.get(0)).nodes.size() * 2);
    	else
    		hn = new HashSet();
    	int i;
    	int floatNum = 0;
    	for(i=0; i<trees.size(); i++) {
    		Tree t = (Tree)trees.get(i);
    		for(int j=0; j<t.nodes.size(); j++) {
    			String n = ((TreeNode)t.nodes.get(j)).getName();
    			if(n.length()>0)
    				hn.add(n);
    		}
    	}
    	i=0;
    	String s[] = new String[hn.size()];
    	Iterator it = hn.iterator();
    	while (it.hasNext()) {
    		s[i] = (String) it.next();
    		i++;
    	}
    	Arrays.sort(s); // one sort instead of making s a treeset which would sort with every add
    	searchFrame.initializeList(new ArrayList(Arrays.asList(s)));
	}

    public int addGroup(boolean thisTreeOnly) {
	int groupCount = groups.size();
	RangeList g = new RangeList(groupCount);
	g.setEnabled(true);
	g.setThisTreeOnly(thisTreeOnly);
	if (AccordionDrawer.debugOutput)
	    System.out.println("adding group " + groupCount);
	groups.add(g);
	groupsByIndex.add(g); // permanent ordering
	return groupCount;
    }

    public void setGroupPriority(int group) {
//    System.out.println("priority change, group " + group + " " + foundGroup);
	Object g = groupsByIndex.get(group);
	if (group <= markGroup.length) {
	    groups.remove(g);
	    groups.add(1,g); // put it first of changeable ones
	    // update keys/alpha for marked colors

	    Iterator iter = groups.listIterator(1);
	    int currKey = 0;

	    while (currKey < markGroup.length)
	    {
	    	RangeList currGroup = (RangeList)iter.next();
//			System.out.println("setting group " + currGroup.getColor().getAlpha() + " " + currGroup.hashCode() + " to key " + currKey);
	    	currGroup.setKey(currKey++);
	    }
	}
    }

	
	public void clearDrawers()
	{
		Iterator atdIter = treeDrawers.iterator();
		if( treeDrawers.size() > 1)
		while( atdIter.hasNext())
		{
			AccordionTreeDrawerFinal a = (AccordionTreeDrawerFinal) atdIter.next();
			
			if( a.baseBox != null && a == a.getBaseBox().getDrawer())
			{
			if( a.getFlashBox() != null )
				a.getFlashBox().undraw();
			a.baseBox = null;
			a.setFlashBox(null);
//			a.requestRedraw(); // too much drawing
			a.actionmodeReset();
			}		
		}
	}

    // deliberately don't have removeGroup method. if you don't want
    // it anymore, disable it. why? we want the index in the groups
    // vector to match the key in the rangelist. maybe there's a less
    // hardline way to do this?...

    // what sort of error checking should we do here if group is out
    // of bounds? be silent, return null where appropriate? print out
    // error message? throw an informative exception? ...

    public void setGroupColor(int group, Color c) {
	((RangeList)groupsByIndex.get(group)).setColor(c);
    }
	
    public void setGroupEnable(int group, boolean on) {
	((RangeList)groupsByIndex.get(group)).setEnabled(on);
    }

    private void ascendTreeMark(AccordionTreeDrawer atd, AccordionTreeDrawer currTree, 
            TreeNode directlyMarked, TreeNode n, RangeList addToGroup)
    {
        // ascend currTree (from n, the bcn of directlymarked, to the root) until no more marking needs to be done
        TreeNode currIndirectNode = n.parent;
//        System.out.print(0);
        if (currIndirectNode == null) return;
//        System.out.print(1);
        TreeNode currDirectNode = TPs.getBestCorrNode(currTree.tree, currIndirectNode, atd.tree, edgeweightLevelUsed);
        if (currDirectNode == null) {return;}
//        System.out.print(2);
//      System.out.println("addedAlready: " + currChild.key + " " + currTree.getKey());
      boolean addedAlready = addToGroup.isThisRangeInList(currIndirectNode.key, currIndirectNode.key, currTree);
//      System.out.println("fitsIntoDirect: " + currDirectNode.key + " " + atd.getKey());
      boolean fitsIntoDirect = addToGroup.isThisRangeInList(currDirectNode.key, currDirectNode.key, atd);
//      System.out.println("added: " + addedAlready + "   fits: " + fitsIntoDirect);
      while (!addedAlready && fitsIntoDirect)
//        while (!addToGroup.isThisRangeInList(currIndirectNode.key, currIndirectNode.key, currTree) &&
//                currDirectNode == directlyMarked)
        {
//          System.out.print('.');
            //System.out.println("ascend mark for " + currIndirectNode);
            addToGroup.addRange(currIndirectNode.key, currIndirectNode.key, currTree);
            currIndirectNode = n.parent;
            if (currIndirectNode == null)
                break;
            currDirectNode = TPs.getBestCorrNode(atd.tree, currIndirectNode, currTree.tree, edgeweightLevelUsed);
            if (currDirectNode == null)
                break;
            
            addedAlready = addToGroup.isThisRangeInList(currIndirectNode.key, currIndirectNode.key, currTree);
            fitsIntoDirect = addToGroup.isThisRangeInList(currDirectNode.key, currDirectNode.key, atd);
        }
//      System.out.println(3);
    }
    
    private void descendTreeMark(AccordionTreeDrawer atd, AccordionTreeDrawer currTree,
            TreeNode directlyMarked, TreeNode n, RangeList addToGroup)
    {
        // descend currTree (from n to leaves recursively), until no more marking is needed
        for (int i = 0; i < n.numberChildren(); i++)
        {
            TreeNode currChild = n.getChild(i);
//            System.out.println("child " + i + " checking for descent: " + currChild);
            TreeNode currDirectNode = TPs.getBestCorrNode(currTree.tree, currChild, atd.tree, edgeweightLevelUsed);
            if (currDirectNode == null) {continue;}
//            System.out.println("addedAlready: " + currChild.key + " " + currTree.getKey());
            boolean addedAlready = addToGroup.isThisRangeInList(currChild.key, currChild.key, currTree);
//            System.out.println("fitsIntoDirect: " + currDirectNode.key + " " + atd.getKey());
            boolean fitsIntoDirect = addToGroup.isThisRangeInList(currDirectNode.key, currDirectNode.key, atd);
//            System.out.println("added: " + addedAlready + "   fits: " + fitsIntoDirect);
            if (!addedAlready && fitsIntoDirect)
            {
//                System.out.println("adding on descent: " + currChild);
                addToGroup.addRange(currChild.key, currChild.key, currTree);
                descendTreeMark(atd, currTree, directlyMarked, currChild, addToGroup);
            }
        }
//        System.out.println("d" + n.numberChildren() + " " + n.key + " RangeList: " + addToGroup);
    }
    
    // add nodes indirectly marked in all trees
    public void addNodesToGroup(int min, int max, int group, AccordionTreeDrawer atd) {
    	RangeList addToGroup = (RangeList)groupsByIndex.get(group); 
    	
		addToGroup.addRange(min, max, atd);
	//System.out.println("range now has " + ((RangeList)groupsByIndex.get(group)).ranges.size() + " elements");
	//System.out.println("added min " + min + " max " + max + " group " + group + " tree " + atd.getKey());
		// must add the nodes from all other trees as well
		Iterator iter = treeDrawers.iterator();
//		System.out.println("RangeList: " + addToGroup);
		while (iter.hasNext())
		{
			AccordionTreeDrawer currTree = (AccordionTreeDrawer)iter.next();
			if (currTree != atd) // ignore the directly marked tree
			{
				for (int i = min; i <= max; i++) // over all directly marked nodes
				{
//				    System.out.println(i + "RangeList: " + addToGroup);
				    TreeNode directlyMarked = atd.getNodeByKey(i);
					TreeNode n = TPs.getBestCorrNode(atd.tree, directlyMarked, currTree.tree, edgeweightLevelUsed);
					if (n != null)
					{
						// the key represents the current subtree selected; all children of the bcn of this subtree are added to the range
						// the subtree selected also needs to be walked to make sure all potential forest created with bcn's is handled
					    if (!addToGroup.isThisRangeInList(n.key, n.key, currTree))
					    {
//					        System.out.println("adding, checking on ascend/descend: " + n);
					        addToGroup.addRange(n.key, n.key, currTree);
							// after getting the corresponding node, search up the tree until nothing is needed to be marked
							ascendTreeMark(atd, currTree, directlyMarked, n, addToGroup);
							// also descend the tree until all nodes are marked that need to be
							descendTreeMark(atd, currTree, directlyMarked, n, addToGroup);

					        if(!stateFrame.LCAMark.isSelected())
					            addToGroup.addRange(n.key, n.key, currTree);
					        else
					            addToGroup.addRange(n.key, n.getMax(), currTree);
					    }
					}
				}
			}
		}
    }

    public void removeNodesFromGroup(int min, int max, int group, AccordionTreeDrawer atd) { 
	((RangeList)groupsByIndex.get(group)).removeRange(min, max, atd);
    }

    public Color getGroupColor(int group) { 
	return ((RangeList)groupsByIndex.get(group)).getColor();
    }

    public void clearGroup(int group) {
    	RangeList groupList = (RangeList)groupsByIndex.get(group);     	
	    groupList.clear();
//	    groupList.setColorTree(null);
	
    }

    public void unmarkGroup(int group) {
    // lcaNode.removeAll(lcaNode);
	RangeList groupList = (RangeList)groupsByIndex.get(group); 
	RangeInTree rit;
		if(groupList.getNumRanges() != 0)
		{
			rit = groupList.getFirst();
			for (int i=0; i<lcaNode.size(); i++)
			{
				if(((TreeNode)lcaNode.get(i)).getKey() == rit.getMin() )
				{
//					System.out.println("lca node key: " + rit.getMin());
				   lcaNode.remove(i);			
				}
			    }
		    }
    if(stateFrame.LCAMark.isSelected())
      this.doLCAGeom(lcaGroup, (AccordionTreeDrawer)treeDrawers.get(0));
	changedMarks();
	clearGroup(group);
    }

    RangeList getGroup(int group) { 
	return ((RangeList)groupsByIndex.get(group));
    }

	int[] seedGroups(AccordionTreeDrawerFinal atd, int[] bboxRange)
	{
		Iterator GroupIter = groups.iterator();
		while (GroupIter.hasNext())
		{
			RangeList group = (RangeList)((RangeList)GroupIter.next()).onlyThisAD(atd);
			if (group.getEnabled())
			{
				Iterator RangeIter = group.getRanges().iterator();
				int count=0;
				while (RangeIter.hasNext())
				{ // seed all ranges for now, add count restriction back later if necessary && count < 5) {
					RangeInTree r = (RangeInTree) RangeIter.next();
					int min = r.getMin(), max = r.getMax();
					TreeNode rootNode;
//					if (group.getKey() == flashGroup)
//					{ // put the block of flash in the queue instead of just the root
//						// flash "group" is only one node, the root of the flash tree
//						if (bboxRange[0] != -1 && bboxRange[1] != -1)
//						{
//							bboxRange[0] = atd.getNodeByKey(min).getCell().getMinLine(AccordionDrawer.Y);
//							bboxRange[1] = atd.getNodeByKey(min).getCell().getMaxLine(AccordionDrawer.Y);
//						}
////						System.out.println(" (qq) " + atd.getToDrawQ());
//					}
//					else
					{
						do
						{
							rootNode = atd.getNodeByKey(min);
							atd.addToDrawQueue(rootNode);
							min = rootNode.rightmostLeaf.key+1;
						}
						while (min <= max);
					}
					count++;
				}
			}
		}
		return bboxRange;
	}
    
	void seedQueue(AccordionTreeDrawerFinal atd) {
		if (AccordionDrawer.debugOutput)
			System.out.println("seeding queue");
		atd.clearQueue();
		InteractionBox bbox = atd.baseBox;
		int firstRange[] = {bbox == null ? -1 : bbox.getMinLine(AccordionDrawer.Y),
				bbox == null ? -1 : bbox.getMaxLine(AccordionDrawer.Y)};

//		RangeList seedRangeList = asd.returnSeedingQueue();
		firstRange = seedGroups(atd, firstRange);
		
		Vector drawQueue = atd.getToDrawQ();
		int numSeededGroups = drawQueue.size();
		Iterator iter = atd.getListOfLeaves().iterator();
		while (iter.hasNext())
		{
			SplitCell sc = (SplitCell)iter.next();
			if (sc.max > firstRange[0] && sc.min < firstRange[1])
				drawQueue.add(numSeededGroups, new RangeInTree(sc.min, sc.max, atd)); // higher priority
			else
				drawQueue.add(new RangeInTree(sc.min, sc.max, atd));
		}
		if (AccordionDrawer.debugOutput)
			System.out.println("seeded: " + drawQueue.size() + " frame: " + atd.getFrameNum());
    	if (atd.basePass || atd.groupPass) // reverse queue, linear but only for taking pictures
    	{
    		Vector temp = new Vector(drawQueue);
    		drawQueue.clear();
    		while (temp.size() > 0)
    			drawQueue.add(temp.remove(temp.size()-1));
    	}
//    	System.out.println(" (QQ) " + atd.getToDrawQ());

	}


//t.getLeafCount()-1

    // built-in group precedence ordering determines which group color
    // wins out. just stop when you find the first one, because order
    // matters.

	ArrayList getColorsForRange(int objMin, int objMax,
	  AccordionTreeDrawer callingTreeDrawer) {
		ArrayList returnList = new ArrayList();
		//System.out.println("getting colors: " + objMin + " " + objMax);
		Tree callingTree = callingTreeDrawer.getTree();

		//	System.out.println("TOP objmin " + objMin + " objmax " + objMax  + " callingtree " + callingTree.getKey());

		RangeInTree fakeRange =
			new RangeInTree(objMin+1, objMax-1, callingTreeDrawer);

		RangeList foundGroup = null;
		Iterator GroupIter = groups.iterator();
		RangeList.returnObject = true;
		while (GroupIter.hasNext()) { // && returnList.size() < 1) {
			RangeList group = (RangeList) GroupIter.next();

			if (group.getEnabled() && group.getRanges().contains(fakeRange)) {
				RangeInTree foundRange = RangeList.matchRange;
				if (foundRange != null && group.getColor() != null)
					returnList.add(group.getColor());
			}
		}
		RangeList.returnObject = false;
//		System.out.println("got colors: " + objMin + " " + objMax + " " + returnList);
		return returnList;
	}
	
    /** 
     * Acts as a toggle on group members: turn off old ones, turn on
     * new ones, do lightweight frontbuffer drawing per item not whole
     * screen redraw
     *
     * @author  Tamara Munzner
     * @see AccordionDrawer.Tree
     * @see AccordionDrawer.TreeNode
     * @see TreeJuxtaposer.RangeList
     *
     */

    void doFlashGeom(CellGeom flashGeom, int group, AccordionTreeDrawerFinal thisTree, int x, int y, boolean all) {
    	//System.out.println("doFlashGeom: " + x + " " + y);
	RangeList theGroup = getGroup(group);
	TreeNode flashNode;
	//System.out.println("fg "+flashGeom+" fgo "+flashGeomOld);
	if (null == flashGeom) {
	    //	    if (null != flashGeomOld) {

	    if (true) {
		// the first time flashGeom goes null, we need to call
		// setFlash with a null geometry to make sure the
		// highlight is turned off.
		flashNode = null;
	    } else return;
	} else {
	    // flashGeom is an edge. get its associated node instead.
	    flashNode = thisTree.getNodeByKey(flashGeom.getKey());
	    //	    	    System.out.println("flash normal");
	}
	//System.out.println("fn "+flashNode+" fgo "+flashGeomOld);
	if (flashNode == flashGeomOld) return;
	// we're still doing the group book-keeping so that we know
	// which tree the item is attached to. we just don't want to
	// enable the group, since then we're left with yellow boxes
	// that aren't easy to clean up with this frontbuffer trick!
	clearGroup(group);
	if (null != flashNode)
	    addNodesToGroup(flashNode.getKey(), flashNode.getKey(), group, thisTree);
	Iterator tdIter = treeDrawers.iterator();
	while (tdIter.hasNext()) {
	    AccordionTreeDrawer atd = (AccordionTreeDrawer)tdIter.next();
	    // we're assuming flashGeom is a TreeNode, since we only put nodes in above (not edges)
	    TreeNode bcn = (null == flashNode) ? null :
			TPs.getBestCorrNode(thisTree.tree, flashNode, atd.tree, edgeweightLevelUsed);
	    boolean sametree = thisTree.tree == atd.tree;
	    if (!sametree)
	    	atd.setFlash((CellGeom)bcn,theGroup.getColor(), -1, -1, sametree);
	    else
	    	atd.setFlash((CellGeom)bcn,theGroup.getColor(), x, y, sametree);
	    if (null == flashNode) { // tell it to turn off
			atd.setFlash(null,theGroup.getColor(), x, y, sametree);
	    }
	}
	//System.out.println("TJ set fgo "+flashNode);
	flashGeomOld = flashNode;
    }

    void doSelectGeoms(ArrayList attachedGeoms, boolean selectSubtree, int group, AccordionTreeDrawer thisTree) {
	clearGroup(group);
	if (null != attachedGeoms) {
	    Iterator geomIter = attachedGeoms.iterator();
	    while (geomIter.hasNext()) {
		CellGeom cg = (CellGeom) geomIter.next();
		doSelectGeom(cg, selectSubtree, group, thisTree);
		System.out.println("this tree node key " + cg.getKey() + " min " + cg.getMin() + " max " + cg.getMax() + " thistree " + thisTree.getKey());
	    }
	}
	requestRedrawAll();
    }

    void doSelectGeom(CellGeom g, boolean selectSubtree, int group, AccordionTreeDrawer thisTree) {
        if (AccordionDrawer.debugOutput)
            System.out.println("key to select: "+g.getKey());
	int which = (true == selectSubtree) ? g.getMax() : g.getKey();
	addNodesToGroup(g.getKey(), which, group, thisTree);
	changedMarks();
	RangeList r = getGroup(group);
//	r.setColorTree(new ColorTree(r));
	//System.out.println("range tree for selected group " + group + " is: " + r.getColorTree());
    }

   void doLCAGeom(int group, AccordionTreeDrawer thisTree)
   {
   	if (true) return;
   	clearGroup(group);
   if(lcaNode.size()<2)
   {
   	 System.out.println("need 2 or more nodes to have a LCA");
   	 return; 
   }
    Iterator it = lcaNode.iterator();
    TreeNode tn1; 
   	tn1= (TreeNode)lcaNode.get(0); 
    TreeNode lca = tn1;
     while(it.hasNext())
      lca = lca.lca((TreeNode)it.next());
    
   for(int i=0; i<lcaNode.size(); i++)
    {
    TreeNode current = (TreeNode)lcaNode.get(i);
     while(current != null && current.getKey() != lca.getKey())
      {
     	addNodesToGroup(current.getKey(), current.getKey(), group, thisTree);
     	current=current.parent;
     }
    }
		addNodesToGroup(lca.getKey(), lca.getKey(), group, thisTree);
	changedMarks();
	RangeList r = getGroup(group);
//	r.setColorTree(new ColorTree(r));
     }

    /**
     * Compute perfectly matched nodes between two trees
     *
     * A node X is perfectly matched by a node Y if they have exactly
     * the same leaf set.
     *
     * @see AccordionDrawer.Tree
     * @see AccordionDrawer.TreeNode
     * @see TreeJuxtaposer.TreePairs
     **/

    void doStructuralDiff(AccordionTreeDrawer atd1, AccordionTreeDrawer atd2, int group) {
	Tree t1 = atd1.getTree();
	Tree t2 = atd2.getTree();
	RangeList groupList = getGroup(group);
	for(int i = 0; i < t1.nodes.size(); i++) {
	    TreeNode n = (TreeNode)t1.nodes.get(i);
	    float score = TPs.getBestCorrNodeScore(t1, n, t2, edgeweightLevelUsed);
	    n.setBcnScore(score);
	    if (score < getBcnScore()) {
		//	    	System.out.println("adding node to diffs: " + n);
		groupList.addRange(n.getKey(), n.getKey(), atd1);
//		addNodesToGroup(n.getKey(), n.getKey(), group, atd1);
	    }
	}
	RangeList diffGroupRangeList = getGroup(diffGroup);
//	diffGroupRangeList.setColorTree(new ColorTree(diffGroupRangeList));
    }

   void setFocus(CellGeom g,  AccordionTreeDrawer atd) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) {
	    AccordionTreeDrawer thisTree = (AccordionTreeDrawer)i.next();
	    TreeNode n = atd.tree.getNodeByKey(g.getKey());
	    if (thisTree != atd) {
		n = TPs.getBestCorrNode(atd.tree, n, thisTree.tree, edgeweightLevelUsed); 
		if (n != null) thisTree.setFocusCell(n.getCell());
		else System.out.println("bad focus");
	    } else {
		thisTree.setFocusCell(n.getCell());
	    }
	}
    }
    
    public String getLabelByFQName(String FQName)
    {
    	String returnValue = null;
    	TreeNode n;
    	Tree t;
    	Iterator i = treeDrawers.iterator();
    	while (i.hasNext())
    	{
    		t = ((AccordionTreeDrawer)i.next()).getTree();
    		if ((n = t.getNodeByName(FQName)) != null)
    		{
    			returnValue = n.label;
    		}
    	}
    	return returnValue;
    }
    
    public void setBigGrid(boolean on) {

    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setBigGrid(on);
    }
    public void toggleBigGrid() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleBigGrid();
    }

    public void setCacheRange(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setCacheRange(on);
    }
    public void toggleCacheRange() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleCacheRange();
    }
    public void setCheckTime(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setCheckTime(on);
    }
    public void toggleCheckTime() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleCheckTime();
    }
    public void setColorGrid(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setColorGrid(on);
    }
    public void toggleColorGrid() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleColorGrid();
    }

    public void setDimBrite(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDimBrite(on);
    }
    public void toggleDimBrite() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDimBrite();
    }
    public boolean getDimBrite() {
      if (treeDrawers.size() > 0)
              return ((AccordionTreeDrawer)treeDrawers.get(0)).dimbrite;
      return false;
    }


    public void setDimColors(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDimColors(on);
    }
    public void toggleDimColors() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDimColors();
    }
    public boolean getDimColors() {
            if (treeDrawers.size() > 0)
                    return ((AccordionTreeDrawer)treeDrawers.get(0)).dimcolors;
            return false;
    }


    public void setDrawLabels(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDrawLabels(on);
    }
    public void toggleDrawLabels() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDrawLabels();
    }
    public boolean getDrawLabels() {
            if (treeDrawers.size() > 0)
                    return ((AccordionTreeDrawer)treeDrawers.get(0)).drawlabels;
            return false;
    }

    public void setDrawOrder(int order) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDrawOrder(order);
    }
    public void toggleDrawOrder() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDrawOrder();
    }

    public void setDumpStats(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDumpStats(on);
    }
    public void toggleDumpStats() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDumpStats();
    }
    public void setGuarVis(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setGuarVis(on);
    }
    public void toggleGuarVis() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleGuarVis();
    }

    public void setExpandLeaves(int expand, boolean redraw) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setExpandLeaves(expand, redraw);
    }
    public void toggleExpandLeaves() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleExpandLeaves();
    }

    public void setLabelBuffer(int buffer) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelBuffer(buffer);
    }
    public void increaseLabelBuffer() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).increaseLabelBuffer();
    }
    public void decreaseLabelBuffer() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).decreaseLabelBuffer();
    }
    public void setLabelBuffer(int buffer, int xy) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelBuffer(buffer, xy);
    }
    public void increaseLabelBuffer(int xy) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).increaseLabelBuffer(xy);
    }
    public void decreaseLabelBuffer(int xy) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).decreaseLabelBuffer(xy);
    }

    public void setLabelDrawBack(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelDrawBack(on);
    }
    public void toggleLabelDrawBack() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleLabelDrawBack();
    }

    public void setLabelPopup(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelPopup(on);
    }
    public void toggleLabelPopup() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleLabelPopup();
    }

    public void setLabelPosRight(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelPosRight(on);
    }
    public void toggleLabelPosRight() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleLabelPosRight();
    }

    public void setLabelTransp(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLabelTransp(on);
    }
    public void toggleLabelTransp() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleLabelTransp();
    }
    public void setDrawSplits(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setDrawSplits(on);
    }
    public void toggleDrawSplits() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleDrawSplits();
    }

    // not called
    public void setLineThickness(int thickness) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLineThickness(thickness);
    }
    
    // no redraw
    public void increaseLineThickness() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).increaseLineThickness();
    }
    public void decreaseLineThickness() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).decreaseLineThickness();
    }

    public void setLinkedNav(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setLinkedNav(on);
    }
    public void toggleLinkedNav() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleLinkedNav();
    }
    public boolean getLinkedNav() {
            if (treeDrawers.size() > 0)
                    return ((AccordionTreeDrawer)treeDrawers.get(0)).linkednav;
            return false;
    }


    public void setMaxFontHeight(int fontsize) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setMaxFontHeight(fontsize);
    }
    public void increaseMaxFontHeight() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).increaseMaxFontHeight();
    }
    public void decreaseMaxFontHeight() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).decreaseMaxFontHeight();
    }

    public void setMinFontHeight(int fontsize) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setMinFontHeight(fontsize);
    }
    public void increaseMinFontHeight() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).increaseMinFontHeight();
    }
    public void decreaseMinFontHeight() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).decreaseMinFontHeight();
    }

    public void setNoFlash(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setNoFlash(on);
    }
    public void toggleNoFlash() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleNoFlash();
    }

    public void setNoGeoms(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setNoGeoms(on);
    }
    public void toggleNoGeoms() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleNoGeoms();
    }

    public void setNoGrid(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setNoGrid(on);
    }
    public void toggleNoGrid() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleNoGrid();
    }

    public void setShowAlg(boolean on) {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).setShowAlg(on);
    }
    public void toggleShowAlg() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).toggleShowAlg();
    }

    public void setShowDiffs(boolean on) {
	showStructDiffs = on;
//	System.out.println("showdiffs " + on);
	setGroupEnable(diffGroup, on);
	changedMarks();
	requestRedrawAll();
    }
    public void toggleShowDiffs() {
	showStructDiffs = !showStructDiffs;
	setGroupEnable(diffGroup, showStructDiffs);
	changedMarks();
	requestRedrawAll();
    }
    public boolean getShowDiffs() {
            return showStructDiffs;
    }
	public void setLCAGroup(boolean on) { stateFrame.setLCAGroup(on); }
	public boolean getLCAGroup() { return stateFrame.getLCAGroup(); }
	public void setLCAMark(boolean on) { stateFrame.setLCAMark(on); }
	public boolean getLCAMark() { return stateFrame.getLCAMark(); }


    public void reset() {
    	Iterator i = treeDrawers.iterator();
	while (i.hasNext()) 
	    ((AccordionTreeDrawer)i.next()).reset();
    }

    /** Return one TreeNode per group range item.
     * @author = Tamara Munzner, Serdar Tasiran
     */
    public ArrayList getGroupForest(int groupindex, AccordionTreeDrawer atd) {
	RangeList group = getGroup(groupindex);
	ArrayList forestRoots = new ArrayList();
	Iterator groupIter = group.getRanges().iterator();
	while (groupIter.hasNext()) {
	    RangeInTree r = (RangeInTree) groupIter.next();
	    // if item is singleton, item.max == item.min
	    AccordionTreeDrawer ratd = r.getTree();
	    Tree thisTree = ratd.tree;
	    int key = r.getMin(); // if it's a range, min is the top node.
	    TreeNode tn = thisTree.getNodeByKey(key);
	    if (thisTree == atd.tree) {
		//System.out.println("tj.rf: " + tn.getKey());
		forestRoots.add(tn);
	    } else if (!group.isThisTreeOnly()) {
	    	// get list of nodes that match subtree rooted by tn
	    	ArrayList list = TPs.getBestNodeList(thisTree, tn, atd.tree, edgeweightLevelUsed);
		//System.out.println("tj.rf: " + tn.getKey() + " corr " + corr.getKey());
		//forestRoots.add(corr);
			if (list == null) // save space in hash, no forest
				forestRoots.add(TPs.getBestCorrNode(thisTree, tn, atd.tree, edgeweightLevelUsed));
			else
			for (int i = 0; i < list.size(); i++)
			{
				forestRoots.add((TreeNode)list.get(i));
	   		}
	    }
	}
	return forestRoots;
    }

    public void resizeGroup(RangeList group, int numSteps, boolean grow, boolean horizontal, boolean vertical, AccordionTreeDrawer atd) {
	{
		atd.endAllTransitions();
		Hashtable newToMove = new Hashtable();
		//		RangeList groupForThisATD = (RangeList)group.onlyThisAD(atd);
		if (horizontal)
		{
			RangeList horizGroup = (RangeList) group.onlyThisAD(atd);
			if (!grow)
			{
				horizGroup = (RangeList) horizGroup.nodeKeyToSplitLine(AccordionDrawer.X, true).flipRangeToShrink(AccordionDrawer.X, atd);
			}
			else
				horizGroup = horizGroup.nodeKeyToSplitLine(AccordionDrawer.X, true);
			atd.getSplitLine(AccordionDrawer.X).resizeForest(horizGroup, numSteps, newToMove, atd.getInflateIncr());
		}
		if (vertical) 
		{
			RangeList vertGroup = (RangeList) group.onlyThisAD(atd);
			if (!grow)
			{
				vertGroup =
					(RangeList) vertGroup.nodeKeyToSplitLine(
						AccordionDrawer.Y, false).flipRangeToShrink(
						AccordionDrawer.Y, atd);
			}
			else
				vertGroup = vertGroup.nodeKeyToSplitLine(AccordionDrawer.Y, false);
			SplitLine splitLineY = atd.getSplitLine(AccordionDrawer.Y);
			RangeList blockList = new RangeList(vertGroup, splitLineY.makePixelRanges((float)1/10), splitLineY);
			atd.getSplitLine(AccordionDrawer.Y).resizeForest(
				blockList,
				numSteps,
				newToMove,
			atd.getInflateIncr());
		}
		atd.toMove = newToMove;
		}
    }
    
	public void resizeGroup (RangeList group, int numSteps, boolean grow, boolean horizontal,
		boolean vertical)
	{
		//RangeList group = getGroup(groupindex);
		Iterator tdIter = treeDrawers.iterator();
		while (tdIter.hasNext())
		{
			AccordionTreeDrawer atd = (AccordionTreeDrawer) tdIter.next();
			resizeGroup(group, numSteps, grow, horizontal, vertical, atd);
		}
	}
    
   
///**
// * Resize (grow=true ==> expand, shrink otherwise) the subtree
// * rooted at node contained in Geoms.
// * 	
// * Geoms is a list of selected items and is expected to contain
// * one tree node or edge.
// *
// * @author   Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
// * 
// *
// * @see      AccordionDrawer.GridCell
// * @see      AccordionDrawer.CellGeom
// * @see      AccordionDrawer.Tree
// * @see      AccordionDrawer.TreeNode
// */
//
//public void resizeSubtree(TreeNode tn, boolean grow, int numSteps, boolean horizontal, boolean vertical)
//{
//	Hashtable newToMove = new Hashtable();
//	
//	RangeList r = new RangeList(99);
//	AccordionTreeDrawer d = (AccordionTreeDrawer)tn.getCell().drawer;
//	r.addRange(tn.getMin(), tn.getRightmostLeaf().getMax(), d);
//	
//	if (horizontal)
//	{
//
//		if (!grow)
//		{
//			r = (RangeList)r.nodeKeyToSplitLine(AccordionDrawer.X).flipRangeToShrink(AccordionDrawer.X, atd);
//		}
//		else
//		r = r.nodeKeyToSplitLine(AccordionDrawer.X);
//		d.getSplitLine(AccordionDrawer.X).resizeForest(r, numSteps, newToMove);
//		}
//			
//	if (vertical)
//	{
//		System.out.println("Do the vertical resize subtree");
//	}
//	
//	
//	
//}


    public void resizeRectOthers(TreeNode tn, double[] changeRatio, int numAnimSteps,
    	AccordionTreeDrawer thisatd) {
    if (null == tn)
    	return;
    
	Iterator tdIter = treeDrawers.iterator();
	while (tdIter.hasNext()) {
	    AccordionTreeDrawer atd = (AccordionTreeDrawer)tdIter.next();
		while (atd.keepMoving())
		{
			if (AccordionDrawer.debugOutput)
				System.out.println(" (rzRO) killing transitions");
			atd.endAllTransitions(); // end transitions, new queue is going to be swapped in
		}
		Hashtable newToMove = new Hashtable();
	    if (thisatd != atd) {
		TreeNode bcn = TPs.getBestCorrNode(thisatd.tree, tn, atd.tree, edgeweightLevelUsed);
		if (null != bcn) {
			TreeNode minLeaf = bcn.leftmostLeaf;
			TreeNode maxLeaf = bcn.rightmostLeaf;
			RangeList range;// = new RangeList[2];
			range = new RangeList(98);
			range.addRange(bcn.getCell().getMinLine(AccordionDrawer.X)+1,
				minLeaf.getCell().getMaxLine(AccordionDrawer.X), atd);
			atd.getSplitLine(AccordionDrawer.X).resizeForest(range, numAnimSteps, newToMove, changeRatio[AccordionDrawer.X]);

			range = new RangeList(99);
			range.addRange(minLeaf.getCell().getMinLine(AccordionDrawer.Y)+1,
				maxLeaf.getCell().getMaxLine(AccordionDrawer.Y), atd);
			atd.getSplitLine(AccordionDrawer.Y).resizeForest(range, numAnimSteps, newToMove, changeRatio[AccordionDrawer.Y]);
			if (atd.debugOutput)
			    System.out.println("incrementing frame from resizeRectOthers");
			atd.incrementFrameNumber(); // so compute place this frame will update, this can be moved to after transitions are created
			
			atd.toMove = newToMove;
		}
	    }
	}
    }

  

//    private void loadNewickTree(String fname) {
	protected void loadNewickTree(String fname) {  
	
	   long start, loaded, processed;
	   if (AccordionDrawer.debugOutput)
	       System.out.println("load0 " + fname);
	   start  = System.currentTimeMillis();
	   Tree t = new Tree();
	   t.load(fname);
	// don't test if the tree already exists, ok to load same tree twice
	//if(getTreeByName(t.getName())!=null) {
	//JOptionPane.showMessageDialog(null, "The tree has already been loaded..", "warning", JOptionPane.WARNING_MESSAGE);
	//  return;
	//}
	   loaded  = System.currentTimeMillis();
	   if (AccordionDrawer.debugOutput)
	       System.out.println("load/parse time " + (loaded-start)/1000.0f + " sec");
	   addTree(t);
	   updateTitle();
	   processed  = System.currentTimeMillis();
	   if (AccordionDrawer.debugOutput)
	       System.out.println("preprocess time " + (processed -loaded)/1000.0f + " sec");
    }

//    private void loadNexusTree(String fname, Vector numbers) {
	protected void loadNexusTree(String fname, Vector numbers) {

	   int cnt = 0;
	   try{
	       FileInputStream in = new FileInputStream(fname);
	       Status.ReInit(in);
	       cnt = parser_status.parseStatus();
	       if (AccordionDrawer.debugOutput)
	           System.out.println("Totally " + cnt + " trees found in file "
			       + fname); 
	       in.close();
	       in = null;
	       System.gc();
	   } catch(Exception e){
	       System.out.println("Error! File " + fname + " not found!");
	       System.exit(1); 
	     };	
       
	if(cnt == 0) {
	    JOptionPane.showMessageDialog(null,  "No tree in the file.", "warning", JOptionPane.WARNING_MESSAGE);
	    return;
	}
	

	if(numbers != null && numbers.size() > 0){
	    for(int i = 0; i < numbers.size(); i++) {
		int index = ((Integer)numbers.elementAt(i)).intValue();
		// tree index should be in the range [1, cnt]
		if(index <= 0 || index > cnt){
		    System.out.println("tree index number should be in the range [1, " + cnt + "], index " + index + " ignored!");
		    continue;
		}
		long start  = System.currentTimeMillis();
		System.out.println("Parse tree " + index + " of file " + fname);
		Tree t = new Tree();
		t.load_nexus(fname, index);
		long loaded  = System.currentTimeMillis();
		if (AccordionDrawer.debugOutput)
		    System.out.println("load/parse time " + (loaded-start)/1000.0f + " sec");
		addTree(t);
		updateTitle();
		long processed  = System.currentTimeMillis();
		if (AccordionDrawer.debugOutput)
		    System.out.println("preprocess time " + (processed -loaded)/1000.0f + " sec");
	    }
	    
	}else{
	
	    NexusTreeLoader ntl = new NexusTreeLoader(this, fname, cnt);
	    if (ntl.enabled)
	    {
	        ntl.setLocation((int) mainFrame.getLocation().getX()+50,
			    (int) mainFrame.getLocation().getY()+100);
	        ntl.pack();
	        ntl.show();
	    }
	    updateTitle();
	 }
    }
    
    public static void main (String[] args) {

	TreeJuxtaposer tj = new TreeJuxtaposer();
	// String title = "TreeJuxtaposer";
	// tj.mainFrame = new JFrame(title);
	// tj.mainFrame.setLayout( new GridLayout(1,2,10,0) );
	
	// get rid of annoying sync warnings
	System.setProperty("java.util.prefs.syncInterval","2000000");

	int i=0;
	AccordionDrawer.loaded = false;
	while (i<args.length) {
		if (args[i].equals("-f")) // flag to fully qualify node names
			AccordionTreeDrawer.fullyQualified = true; // default is false
	    else if (args[i].equals("-nostructdiff")) {
		tj.setShowDiffs(false);
	    } else if (args[i].equals("-matrix")) {
		tj.matrix = true;
	    } else if (args[i].equals("-L")) {
		tj.edgeweightLevels = (new Integer(args[++i])).intValue();
		tj.edgeweightLevelUsed = (new Integer(args[++i])).intValue();
		if(tj.edgeweightLevelUsed>=tj.edgeweightLevels)
		    tj.edgeweightLevelUsed = tj.edgeweightLevels-1;
		else if(tj.edgeweightLevelUsed<0)
		    tj.edgeweightLevelUsed = 0;
	    } else if (args[i].equals("-x")) {
		i++;
		String filename = args[i++];
		Vector numbers = new Vector();
		while(i < args.length){
		    String s = args[i];
		    Integer n;
		    try{
			n = Integer.valueOf(s);
		    }catch(NumberFormatException ne){
			break;
		    }
		    numbers.add(n);
		    i++;
		}
		tj.loadNexusTree(filename, numbers);

		// to offset the next operation i++
		i--;  
		
	    } else {

		tj.loadNewickTree(args[i]);
		tj.treeArgs = args.length-i;
	    }
	    i++;
	}
	
	
//	System.out.println("levels: "+tj.edgeweightLevels+" level used: "+tj.edgeweightLevelUsed);
    if (i == 0)    	tj.mainFrame.setSize(300,300);
//    else    	    tj.mainFrame.pack();
    tj.updateTitle();

    tj.mainFrame.setVisible(true);
    
//    tj.requestRedrawAll(); // now dispatched by setvisible PaintEvent
   }

	public void observe()
		{
			//System.out.println("observing...");
			if (treeDrawers.size() == 0)
				return;
			observeActionGroup();
			observeActionMode();
			observeActionTarget();
			observeMarkGroup();
			stateFrame.observe();
		}
	
	   /** 
		* a more compact version of implementation for observing functions 
		* @author  Jeffrey Zhang
		*
		*/ 
		public void observeActionMode()
		{
			Iterator adIter = treeDrawers.iterator();
			String output = "";
			while( adIter.hasNext())
			{
				
				int treeDrawerActionMode = ((AccordionTreeDrawerFinal)adIter.next()).getActionMode();
				output += DebugFrame.actionModeLabel[treeDrawerActionMode] + ", ";
			}
			debugFrame.result[DebugFrame.ACT_MODE].setText(output);
		
		}

/**
 * 
 * @author 
 *
 */
	public void clearObserve(){
			debugFrame.result[DebugFrame.ACT_MODE].setText(" ");
			debugFrame.result[DebugFrame.ACT_TYPE].setText(" ");
			debugFrame.result[DebugFrame.ACT_GROWDIR].setText(" ");			
			debugFrame.result[DebugFrame.ACT_SELECTION].setText(" ");
			debugFrame.result[DebugFrame.M_GROUP].setText(" ");
		}

		public void observeActionGroup()
		{
			
			int treeDrawerActionGroup = ((AccordionTreeDrawerFinal)treeDrawers.get(0)).getActionGroup();
            debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[treeDrawerActionGroup]);
		}
		
		public void observeActionTarget()
		{
			// action target split into: selectionResolution and growDirection
			int treeDrawerSelectionResolution = ((AccordionTreeDrawerFinal)treeDrawers.get(0)).getSelectionResolution();
			int treeDrawerGrowDirection = ((AccordionTreeDrawerFinal)treeDrawers.get(0)).getGrowDirection();
            debugFrame.result[DebugFrame.ACT_GROWDIR].setText(DebugFrame.actionTargetLabel[treeDrawerGrowDirection]);			
            debugFrame.result[DebugFrame.ACT_SELECTION].setText(DebugFrame.actionTargetLabel[treeDrawerSelectionResolution + 3]);
		}
		
	public void observeMarkGroup()
	{
		int treeDrawerMarkGroup = ((AccordionTreeDrawerFinal)treeDrawers.get(0)).getMarkGroup();
        debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[treeDrawerMarkGroup]);

	}

	public void setBcnScore(float score)
	{
		bcnScore = score;
	}
	public float getBcnScore()
	{
		return bcnScore;
	}
	
//	public void setLeftRight(int i) {leftRight=i; }
//	public int getLeftRight() { return leftRight; }
	
	public void drawRange(RangeList range)
	{
		System.out.println("Drawing range: " + range + " on demand");
		Iterator rangeIter = range.getRanges().iterator();
		while (rangeIter.hasNext())
		{
			RangeInTree rit = (RangeInTree)rangeIter.next();
			AccordionTreeDrawer d = rit.getTree();
			System.out.println("incrementing frame from drawRange");
			d.incrementFrameNumber();
			d.drawRange(d.getNodeByKey(rit.getMin()).leftmostLeaf.getCell().getMinLine(AccordionDrawer.Y) - 1,
						d.getNodeByKey(rit.getMax()).rightmostLeaf.getCell().getMaxLine(AccordionDrawer.Y));
		}
	}

//	public void windowClosing(WindowEvent event) {
//		System.out.println("window closing");
//		System.exit(0);
//	}
//	public void windowClosed(WindowEvent event) {
//		System.out.println("window closed");
//		System.exit(0);
//	}
//	public void windowActivated(WindowEvent event) {
//		System.out.println("window activated");
//	}
//	public void windowDeiconified(WindowEvent event) {
//		System.out.println("window deiconified");
////		requestRedrawAll();
//	}
//	public void windowDeactivated(WindowEvent event) {
//		System.out.println("window deactivated");
//	}
//	public void windowIconified(WindowEvent event) {
//		System.out.println("window iconified");
//	}
//	public void windowOpened(WindowEvent event) {
//		System.out.println("window opened");
//		requestRedrawAll();
//	}
		
	public void doSnapshot()
	{
		Iterator iter = treeDrawers.iterator();
		while (iter.hasNext())
		{
			AccordionTreeDrawer atd = (AccordionTreeDrawer)iter.next();
			atd.takeSnapshot = true; // unset in the drawEnd call
			// these are set one-by-one for 3 drawing cycles
			atd.basePass = true;  // draw the base image, nothing marked, no text
			atd.groupPass = false; // draw the groups, no text
			atd.textPass = false; // draw the text
			try
			{
				atd.newSnapshotName(mainFrame.getTitle(), ui.fc.getCurrentDirectory().getCanonicalPath());
			}
			catch (IOException e)
			{
				System.out.println("Unable to get current directory from file chooser, using home directory");
				atd.newSnapshotName(mainFrame.getTitle(), System.getProperty("user.home"));
			}
			atd.requestRedraw();
		}
		if (AccordionDrawer.debugOutput)
			System.out.println("done the base pass");
//		requestRedrawAll(); // base pass
		
		iter = treeDrawers.iterator();
		while (iter.hasNext())
		{
			AccordionTreeDrawer atd = (AccordionTreeDrawer)iter.next();
			atd.basePass = false;
			atd.groupPass = true;
			atd.requestRedraw();
		}
		if (AccordionDrawer.debugOutput)
			System.out.println("done the group pass");
//		requestRedrawAll(); // group pass
		
		iter = treeDrawers.iterator();
		while (iter.hasNext())
		{
			AccordionTreeDrawer atd = (AccordionTreeDrawer)iter.next();
			atd.groupPass = false;
			atd.textPass = true;
			atd.redraw(); // TODO: does nothing, fix this
		}
		if (AccordionDrawer.debugOutput)
			System.out.println("done the text pass");
//		requestRedrawAll(); // text pass

	}

	/* (non-Javadoc)
	 * @see java.awt.event.WindowFocusListener#windowGainedFocus(java.awt.event.WindowEvent)
	 */
//	public void windowGainedFocus(WindowEvent e) {
//		System.out.println("window gained focus");
//		
//	}
//
//	/* (non-Javadoc)
//	 * @see java.awt.event.WindowFocusListener#windowLostFocus(java.awt.event.WindowEvent)
//	 */
//	public void windowLostFocus(WindowEvent e) {
//		System.out.println("window lost focus");
//	}
	
//	tj.setProgressiveOn(progressiveOn.isSelected());
	public void setProgressiveOn(boolean on)
	{
		Iterator iter = treeDrawers.iterator();
		while (iter.hasNext())
		{
			AccordionTreeDrawerFinal atd = (AccordionTreeDrawerFinal)iter.next();
			atd.ignoreProgressive = !on;
		}
	}
	
};
